/* dt_hash.c
**
** data type for generic hash table
** dtlib-lite library
**
** Copyright (C) 2002, 2003, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.02.04
** ===
*/
#include <stdlib.h>
#include <assert.h>

/* our headers: */
#include "dt_hash.h"


/* internal object declarations: */

struct hash_item {
    void               *key;
    void               *data;
    struct hash_item   *next;
};


struct hash_ {
    struct hash_item   **table;
    /* table stats: */
    int                  tblsize;
    int                  nitems;
    int                  chains;
    int                  collides;
    /* installed hash and compare functions: */
    unsigned int       (*hash_func)(const void *key);
    int                (*key_compare)(const void *key1, const void *key2);
};
    


/* prototypes: */
static
struct hash_item *
new_hash_item(void *key, void *data, struct hash_item *next);


/* HASH METHODS: */

/* XXX
** according to info gcc, inline functions referred to by
** the function's address can't be inlined (??)
*/

static
int
hash_get_tblsize(struct hash *H);

static
int
hash_get_nitems(struct hash *H);

static
int
hash_is_empty(struct hash *H);

static
float
hash_get_load(struct hash *H);

static
int
hash_get_chains(struct hash *H);

static
int
hash_get_collides(struct hash *H);

static
unsigned int
hash_hash_check(struct hash *H, const void *key);

static
void *
hash_insert(struct hash *H, const void *key, void *data);

static
void *
hash_lookup(struct hash *H, const void *key);

static
int
hash_exists(struct hash *H, const void *key);

static
void *
hash_remove(struct hash *H, const void *key);

static
int
hash_retrieve
(
    struct hash  *H,
    void         *key,
    void        **rkey,
    void        **rdata
);

static
int
hash_extract
(
    struct hash  *H,
    void         *key,
    void        **rkey,
    void        **rdata
);

static
void
hash_apply
(
    struct hash  *H,
    void  (*visitor)(const void *key, void *data, void *xtra),
    void  *xtra
);

static
void
hash_clear
(
    struct hash  *H,
    void  (*free_item)(void *key, void *data)
);

static
void
rehash_insert_item(struct hash_ *hash, struct hash_item *item);

static
struct hash *
hash_rehash
(
    struct hash   *H,
    int            tblsize
);

static
void
hash_free(struct hash *H);



/* function definitions: */

/* HASH ITEMS: */
static
struct hash_item *
new_hash_item(void *key, void *data, struct hash_item *next)
{
    struct hash_item *item =
        (struct hash_item *)malloc(sizeof(struct hash_item));

    if(item == NULL) return NULL;

    item->key  = key;
    item->data = data;
    item->next = next;

    return item;
}/*end new_hash_item()*/



/* HASH CONSTRUCTOR: */
struct hash *
new_hash
(
    int tblsize,
    unsigned int (*hash_func)(const void *key),
    int (*key_compare)(const void *key1, const void *key2)
)
{
    struct hash         *H = NULL;
    struct hash_        *hash = NULL;
    struct hash_item   **table = NULL;
    int                  i = 0;

    if(tblsize <= 0) return NULL;
    if((hash_func == NULL) || (key_compare == NULL)) return NULL;

    H = (struct hash *)malloc(sizeof(struct hash));
    if(H == NULL) return NULL;

    hash = (struct hash_ *)malloc(sizeof(struct hash_));
    if(hash == NULL){
        free(H);
        return NULL;
    }

    table = (struct hash_item **)malloc(sizeof(struct hash_item *) * tblsize);
    if(table == NULL){
        free(hash);
        free(H);
        return NULL;
    }

    /* initialize table: */
    for(i = 0; i < tblsize; i++){
        table[i] = NULL;
    }

    /* initialize hash_: */
    hash->table = table;
    hash->tblsize = tblsize;
    hash->nitems = 0;
    hash->chains = 0;
    hash->collides = 0;
    hash->hash_func = hash_func;
    hash->key_compare = key_compare;

    /* initialize Hash: */
    H->hash = hash;

    /* methods: */
    H->size       = &hash_get_tblsize;
    H->items      = &hash_get_nitems;
    H->is_empty   = &hash_is_empty;
    H->load       = &hash_get_load;
    H->chains     = &hash_get_chains;
    H->collides   = &hash_get_collides;
    H->hash_check = &hash_hash_check;
    H->insert     = &hash_insert;
    H->lookup     = &hash_lookup;
    H->exists     = &hash_exists;
    H->remove     = &hash_remove;
    H->retrieve   = &hash_retrieve;
    H->extract    = &hash_extract;
    H->apply      = &hash_apply;
    H->clear      = &hash_clear;
    H->rehash     = &hash_rehash;


    /* destructor: */
    H->free       = &hash_free;

    return H;
}/*end new_hash()*/


/* HASH METHODS: */
static
int
hash_get_tblsize(struct hash *H)
{
    return H->hash->tblsize;
}/*end hash_get_tblsize()*/



static
int
hash_get_nitems(struct hash *H)
{
    return H->hash->nitems;
}/*end hash_get_nitems()*/



static
int
hash_is_empty(struct hash *H)
{
    return (H->hash->nitems == 0);
}/* end hash_is_empty()*/



static
float
hash_get_load(struct hash *H)
{
    return ((float)H->hash->nitems / (float)H->hash->tblsize);
}/*end hash_get_nitems()*/



static
int
hash_get_chains(struct hash *H)
{
    return H->hash->chains;
}/*end hash_get_chains()*/



static
int
hash_get_collides(struct hash *H)
{
    return H->hash->collides;
}/*end hash_get_collides()*/



static
unsigned int
hash_hash_check(struct hash *H, const void *key)
{
    return H->hash->hash_func(key);
}/*end hash_hash()*/



/*if replacing, returns old data; else returns data inserted*/
static
void *
hash_insert(struct hash *H, const void *key, void *data)
{
    struct hash_       *hash  = H->hash;
    struct hash_item  **table = hash->table;
    struct hash_item   *item  = NULL;
    unsigned int         h    = 0;

    /* NULL key is an unchecked error */

    /* compute hash for table index: */
    h = hash->hash_func(key);
    h %= hash->tblsize;

    /* if slot empty, push in the new item: */
    if(table[h] == NULL){
        table[h] = new_hash_item((void *)key, data, NULL);
        hash->nitems++;
        hash->chains++;
        return data;
    }

    /* key already present? */
    item = table[h];
    while(item){
        if(hash->key_compare(key, item->key) == 0){
            /* key already present; replace data: */
            void *olddata = item->data;
            item->data = data;
            return olddata;
        }
        item = item->next;
    }

    /* if here, a key collision; push in new item at head of list: */
    table[h] = new_hash_item((void *)key, data, table[h]);
    hash->nitems++;
    hash->collides++;

    return data;
}/*end hash_insert()*/



static
void *
hash_lookup(struct hash *H, const void *key)
{
    struct hash_       *hash  = H->hash;
    struct hash_item  **table = hash->table;
    struct hash_item   *item  = NULL;
    unsigned int        h     = 0;

    /* NULL key is an unchecked error */

    /* compute hash for table index: */
    h = hash->hash_func(key);
    h %= hash->tblsize;

    item = table[h];
    while(item){
        if(hash->key_compare(key, item->key) == 0){
            return item->data;
        }
        item = item->next;
    }

    /* not found: */
    return NULL;
}/*end hash_lookup()*/



static
int
hash_exists(struct hash *H, const void *key)
{
    struct hash_       *hash  = H->hash;
    struct hash_item  **table = hash->table;
    struct hash_item   *item  = NULL;
    unsigned int        h     = 0;

    /* NULL key is an unchecked error */

    /* compute hash for table index: */
    h = hash->hash_func(key);
    h %= hash->tblsize;

    item = table[h];
    while(item){
        if(hash->key_compare(key, item->key) == 0){
            return 1; 
        }
        item = item->next;
    }

    /* not found: */
    return 0;
}/*end hash_exists()*/



static
void *
hash_remove(struct hash *H, const void *key)
{
    struct hash_       *hash  = H->hash;
    struct hash_item  **table = hash->table;
    struct hash_item   *item  = NULL, *prev = NULL;
    unsigned int        h     = 0;

    /* NULL key is an unchecked error */

    /* compute hash for table index: */
    h = hash->hash_func(key);
    h %= hash->tblsize;

    prev = NULL;
    item = table[h];
    while(item){
        if(hash->key_compare(key, item->key) == 0){
            /* match found; save the data: */
            void  *rdata = item->data;

            /* reset link pointers: */
            if(prev == NULL){
                /* deleting list head: */
                table[h] = item->next;
            }
            else{
                /* deleting within list: */
                prev->next = item->next;
            }

            free(item);

            /* update hash statistics: */
            if(table[h] == NULL){
                /* this slot had only one item, and is now emptied: */
                hash->chains--;
            }
            else{
                /* this slot still has items, so we reduced a collision: */
                hash->collides--;
            }
            hash->nitems--;

            return rdata;
        }
        prev = item;
        item = item->next;
    }

    /* key not found! */
    return NULL;
}/*end hash_remove()*/



static
int
hash_retrieve
(
    struct hash  *H,
    void         *key,
    void        **rkey,
    void        **rdata
)
{
    struct hash_       *hash  = H->hash;
    struct hash_item  **table = hash->table;
    struct hash_item   *item  = NULL;
    unsigned int        h     = 0;

    /* NULL key is an unchecked error */

    /* compute hash for table index: */
    h = hash->hash_func(key);
    h %= hash->tblsize;

    item = table[h];
    while(item){
        if(hash->key_compare(key, item->key) == 0){
            if(rkey)  *rkey = item->key;
            if(rdata) *rdata = item->data;
            return 1; 
        }
        item = item->next;
    }

    /* not found: */
    return 0;
}/*end hash_retrieve()*/



static
int
hash_extract
(
    struct hash  *H,
    void         *key,
    void        **rkey,
    void        **rdata
)
{
    struct hash_       *hash  = H->hash;
    struct hash_item  **table = hash->table;
    struct hash_item   *item  = NULL, *prev = NULL;
    unsigned int        h     = 0;

    /* NULL key is an unchecked error */

    /* compute hash for table index: */
    h = hash->hash_func(key);
    h %= hash->tblsize;

    prev = NULL;
    item = table[h];
    while(item){
        if(hash->key_compare(key, item->key) == 0){
            /* match found; set the key/data for return: */
            if(rkey)  *rkey = item->key;
            if(rdata) *rdata = item->data;

            /* reset link pointers: */
            if(prev == NULL){
                /* deleting list head: */
                table[h] = item->next;
            }
            else{
                /* deleting within list: */
                prev->next = item->next;
            }

            free(item);

            /* update hash statistics: */
            if(table[h] == NULL){
                /* this slot had only one item, and is now emptied: */
                hash->chains--;
            }
            else{
                /* this slot still has items, so we reduced a collision: */
                hash->collides--;
            }
            hash->nitems--;

            return 1;
        }
        prev = item;
        item = item->next;
    }

    /* key not found! */
    return 0;
}/*end hash_extract()*/



static
void
hash_apply
(
    struct hash  *H,
    void  (*visitor)(const void *key, void *data, void *xtra),
    void  *xtra
)
{
    struct hash_      *hash = H->hash;
    struct hash_item  *item = NULL;
    int                slot = 0;

    for(slot = 0; slot < hash->tblsize; slot++){
        item = hash->table[slot];
        while(item){
            visitor((const void *)item->key, item->data, xtra);
            item = item->next;
        }
    } 

    return;
}/*end hash_apply()*/



/* apply user-defined key/data free to items and free all items */
static
void
hash_clear
(
    struct hash  *H,
    void  (*free_item)(void *key, void *data)
)
{
    struct hash_      *hash = H->hash;
    struct hash_item  *item = NULL, *next = NULL;
    int                slot = 0;


    for(slot = 0; slot < hash->tblsize; slot++){
        item = hash->table[slot];
        while(item){
            next = item->next;
            free_item(item->key, item->data);
            free(item);
            item = next;
        }
        hash->table[slot] = NULL;
    }

    hash->nitems = 0;
    hash->chains = 0;
    hash->collides = 0;

    return;
}/*end hash_clear()*/



static
void
rehash_insert_item(struct hash_ *hash, struct hash_item *item)
{
    struct hash_item  **table = hash->table;
    unsigned int        h = 0;

    /* NULL item->key is an unchecked error */

    /* compute hash for table index: */
    h = hash->hash_func(item->key);
    h %= hash->tblsize;

    /* XXX key already present? 
    **
    ** this is why we don't enable changing key_compare() and
    ** hash_func() in hash->rehash(); we assume the original
    ** key_compare() already filtered duplicate keys
    **
    ** client program should create a new hash if it needs to
    ** install a different key_compare()/hash_func()
    */

    if(table[h] == NULL){
        /* if slot empty, push in the new item: */
        item->next = NULL;
        table[h] = item;
        hash->chains++;
    }
    else{
        /* if here, a key collision; push in new item at head of list: */
        item->next = table[h];
        table[h] = item;
        hash->collides++;
    }
    hash->nitems++;

    return;
}/*end rehash_insert_item()*/



static
struct hash *
hash_rehash
(
    struct hash   *H,
    int            tblsize
)
{
    struct hash_        *hash = H->hash, *newhash = NULL;
    struct hash_item   **newtable = NULL, *item = NULL, *next = NULL;
    int                  slot = 0;

    if(tblsize <= 0) return NULL;

    newhash = (struct hash_ *)malloc(sizeof(struct hash_));
    if(newhash == NULL) return NULL;

    newtable = (struct hash_item **)malloc(sizeof(struct hash_item *) * tblsize);
    if(newtable == NULL){
        free(newhash);
        return NULL;
    }

    /* initialize new hash table: */
    for(slot = 0; slot < tblsize; slot++){
        newtable[slot] = NULL;
    }

    /* initialize newhash: */
    newhash->table = newtable;
    newhash->tblsize = tblsize;
    newhash->nitems = 0;
    newhash->chains = 0;
    newhash->collides = 0;
    newhash->hash_func = hash->hash_func;
    newhash->key_compare = hash->key_compare;

    /* walk through the table and rehash existing items: */
    for(slot = 0; slot < hash->tblsize; slot++){
        item = hash->table[slot];
        while(item){
            next = item->next;
            item->next = NULL;
            rehash_insert_item(newhash, item);
            item = next;
        }
    } 

    /* this should be true: */
    assert(hash->nitems == newhash->nitems);

    /* free the old table/hash, replace with new: */
    free(hash->table);
    free(hash);
    H->hash = newhash;

    return H;
}/*end hash_rehash()*/



/* HASH DESTRUCTOR: */
/*
** remember:
**   this only destroys the hash structure,
**   _NOT_ the data within the hash
*/
static
void
hash_free(struct hash *H)
{
    free(H->hash->table);
    free(H->hash);
    free(H);

    return;
}/*end hash_free()*/



/* that's all, folks! */
