/* dt_list.c
**
** data type for generic linked list
** dtlib-lite library
**
** Copyright (C) 2002, 2003, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.02.04
** ===
*/
#include <stdlib.h>
#include <assert.h>

/* our headers: */
#include "dt_list.h"


/* internal declarations: */

struct list_item {
    void               *data;
    struct list_item   *next;
};


struct list_ {
    int                  nitems;
    struct list_item    *head;
    struct list_item    *tail;
    /* installed comparison function: */
    int    (*data_compare)(const void *key1, const void *key2);
};
    


/* prototypes: */
static
struct list_item *
new_list_item(void *data, struct list_item *next);

/* LIST METHODS: */

/* XXX
** according to info gcc, inline functions referred to by
** the function's address can't be inlined
*/

static
int
list_get_nitems(struct list *L);

static
int
list_is_empty(struct list *L);

static
void
list_push(struct list *L, void *data);

static
void *
list_peek(struct list *L);

static
void *
list_pop(struct list *L);

static
void
list_addend(struct list *L, void *data);

static
void
list_reverse(struct list *L);

static
void *
list_search
(
    struct list *L,
    int  (*compare)(const void *key1, const void *key2),
    const void *key
);

static
void
list_apply
(
    struct list  *L,
    void  (*visitor)(void *data, void *xtra),
    void  *xtra
);


static
void
list_clear
(
    struct list  *L,
    void  (*free_item)(void *data)
);


static
void
list_free(struct list *L);



/* function definitions: */

/* LIST ITEMS: */
static
struct list_item *
new_list_item(void *data, struct list_item *next)
{
    struct list_item *item =
        (struct list_item *)malloc(sizeof(struct list_item));

    if(item == NULL) return NULL;

    item->data = data;
    item->next = next;

    return item;
}/*end new_list_item()*/



/* LIST CONSTRUCTOR: */
struct list *
new_list(void)
{
    struct list    *L = NULL;
    struct list_   *list = NULL;

    L = (struct list *)malloc(sizeof(struct list));
    if(L == NULL) return NULL;

    list = (struct list_ *)malloc(sizeof(struct list_));
    if(list == NULL){
        free(L);
        return NULL;
    }

    list->tail = new_list_item(NULL, NULL);
    if(list->tail == NULL){
        free(list);
        free(L);
        return NULL;
    }

    list->head = list->tail;
    list->nitems = 0;
    L->list = list;

    /* methods: */

    L->items    = &list_get_nitems;
    L->is_empty = &list_is_empty;
    L->push     = &list_push;
    L->peek     = &list_peek;
    L->pop      = &list_pop;
    L->addend   = &list_addend;
    L->reverse  = &list_reverse;
    L->search   = &list_search;
    L->apply    = &list_apply;
    L->clear    = &list_clear;
    /* ... */
    L->free     = &list_free;

    return L;
}/*end new_list()*/


/* LIST METHODS: */
static
int
list_get_nitems(struct list *L)
{
    return L->list->nitems;
}/*end list_get_nitems()*/



static
int
list_is_empty(struct list *L)
{
    return (L->list->nitems == 0);
}/* end list_is_empty()*/



static
void
list_push(struct list *L, void *data)
{
    /* XXX bad; no fault if malloc error! */
    L->list->head = new_list_item(data, L->list->head);
    L->list->nitems++;

    return;
}/*end list_push()*/



static
void *
list_peek(struct list *L)
{
    return L->list->head->data;
}/*end list_peek()*/



static
void *
list_pop(struct list *L)
{
    struct list_item   *next = NULL;
    void               *data = NULL;

    if(L->list->head != L->list->tail){
        data = L->list->head->data;
        next = L->list->head->next;
        free(L->list->head);
        L->list->head = next;
        L->list->nitems--;
    }

    return data;
}/*end list_pop()*/
 

static
void
list_addend(struct list *L, void *data)
{
    /* copy new data into former tail: */
    L->list->tail->data = data;
    /* tack on a new null tail: */
    /* XXX no fault if malloc error!! */
    L->list->tail = L->list->tail->next = new_list_item(NULL, NULL);

    L->list->nitems++;

    return;
}/*end list_addend()*/


static
void
list_reverse(struct list *L)
{
    struct list_      *list = L->list;
    struct list_item  *ip = NULL, *tp = NULL;

    if(list->head == list->tail)
        return;

    /* initialize item walker: */
    ip = list->head->next;
    /* set tail sentinal pointer on what will become last element: */
    list->head->next = list->tail;

    /* walk down the list: */
    while(ip != list->tail){
        /* save the current head: */
        tp = list->head;
        /* push in item walker as new head: */
        list->head = ip;
        /* increment item walker here! */
        ip = ip->next;
        /* set new head's next to previous head: */
        list->head->next = tp;
    }
    
    return;
}/*end list_reverse()*/



static
void *
list_search
(
    struct list *L,
    int  (*compare)(const void *key1, const void *key2),
    const void *key
)
{
    struct list_item  *item = L->list->head;

    while(item != L->list->tail){
        if(compare(item->data, key) == 0){
            return item->data;
        }
        item = item->next;
    }

    /* not found: */
    return NULL;

}/*end list_search()*/



static
void
list_apply
(
    struct list *L,
    void (*visitor)(void *data, void *xtra),
    void *xtra
)
{
    struct list_item   *item = L->list->head;

    while(item != L->list->tail){
        visitor(item->data, xtra);
        item = item->next;
    }

    return;
}/*end list_apply()*/



/* LIST CLEAR: */
/*
** applies user-supplied function to free data
** within list items, resulting in empty list
**
** the list itself remains, 0 items, with
** its null sentinal tail
*/
static
void
list_clear
(
    struct list  *L,
    void  (*free_item)(void *data)
)
{
    struct list_item  *item = L->list->head;
    struct list_item  *next = NULL;

    while(item != L->list->tail){
        next = item->next;
        free_item(item->data);
        free(item);
        L->list->nitems--;
        item = next;
    }

    return;
}/*end list_clear()*/



/* LIST DESTRUCTOR: */
/*
** remember:
**   this only destroys the list structure,
**   _NOT_ the data within the list 
*/
static
void
list_free(struct list *L)
{
    struct list_item  *item = L->list->head;
    struct list_item  *next = NULL;

    while(item != L->list->tail){
        next = item->next;
        free(item);
        item = next;
    }

    free(L->list->tail);
    free(L->list);
    free(L);

    return;
}/*end list_free()*/



/* that's all, folks! */
