/* xpin.c
**
** XpinDoc
** an XML parse instance document processing utility
**
** Copyright (C) 2002, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.09.30
** ===
*/
#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* strdup() */

/* this version uses the expat parser: */
#include <expat.h>

/* our headers: */
#include "xpindoc.h"
#include "xpin_event.h"
#include "xpin_node.h"
#include "xpin_pi.h"
#include "xpin_handlers.h"
#include "xpin_writer.h"
#include "xpin_error.h"
#include "xpin.h"

/* XXX get system pagesize */
#define PARSE_BUFFER_SIZE  8*512

/* construct error messages: */
#define MSGBUF_SIZE  512


struct xpin_ns_name {
    char *local_name;
    char *ns_uri;
    char *ns_prefix;
    char *qname;
};


/* prototypes for functions in this module: */
/* xpin methods, to wrap expat parse info: */
static
int
xpin_get_lineno(struct xpin *);

static
int
xpin_get_column(struct xpin *);

static
int
xpin_get_size(struct xpin *);

static
long int
xpin_get_offset(struct xpin *);

static
int
xpin_expat_ns3_parse(struct xpin_ns_name *N, const char *name, const char ns_sep);

static
void
xpin_ns_name_clear(struct xpin_ns_name *N);

static
void
xpin_expat_start_element_handler
(
    struct xpindoc    *X,
    const Xpin_Char   *name,
    const Xpin_Char  **atts
);

static
void
xpin_expat_end_element_handler
(
    struct xpindoc    *X,
    const Xpin_Char   *name
);

/* xpindoc methods: */
static
int
xpindoc_ns_parser(struct xpindoc *);

static
void
xpindoc_set_mydata(struct xpindoc *, void *);

static
void *
xpindoc_get_mydata(struct xpindoc *);

static
int
xpindoc_set_datamode(struct xpindoc *, int);

static
int
xpindoc_get_datamode(struct xpindoc *);

static
int
xpindoc_expat_parse_stream
(
    struct xpindoc *,
    FILE           *
);

static
void
xpindoc_ignore_on(struct xpindoc *);

static
void
xpindoc_ignore_off(struct xpindoc *);

static
int
xpindoc_is_ignored(struct xpindoc *);

static
void
xpindoc_halt_parse(struct xpindoc *);

static
int
xpindoc_is_halted(struct xpindoc *);

static
int
xpindoc_is_croaked(struct xpindoc *);

static
int
xpindoc_set_filter
(
    struct xpindoc *,
    void (*)(struct xpindoc *, const char *, int)
);

static
struct xpin_event *
xpindoc_get_event(struct xpindoc *);

static
enum xpin_event_type
xpindoc_get_event_type(struct xpindoc *);

static
struct xpin_node *
xpindoc_get_node(struct xpindoc *);

static
struct xpin_pi *
xpindoc_get_pi(struct xpindoc *);

static
struct xpin_error *
xpindoc_get_error(struct xpindoc *);

static
int
xpindoc_get_node_depth(struct xpindoc *);

static
const char *
xpindoc_get_node_tagname(struct xpindoc *);

static
const char *
xpindoc_get_node_tagpath(struct xpindoc *);

static
const char *
xpindoc_get_node_content(struct xpindoc *);

static
struct xpin_text *
xpindoc_get_node_text(struct xpindoc *);

static
int
xpindoc_push_output(struct xpindoc *, void *dest, int dest_type);

static
void *
xpindoc_pop_output(struct xpindoc *);

static
int
xpindoc_output_char(struct xpindoc *, char c);

static
int
xpindoc_output_string(struct xpindoc *, const char *string);

static
int
xpindoc_output_chardata(struct xpindoc *, const char *data, int len);
 
static
int
xpindoc_output_printf(struct xpindoc *, const char *format, ...);
 
static
void
free_xpindoc(struct xpindoc *);



/* definitions: */
/* expat compatibility layer: */
/* translate expat namespace before call to meta_handlers: */
static
int
xpin_expat_ns3_parse
(
    struct xpin_ns_name *N,
    const char          *name,
    const char           ns_sep
)
{
    char               *s0 = NULL, *sN = NULL;
    char               *ss[3];
    int                 ix = 0;
    int                 sslen = 0;
    char               *tmp = NULL;

    /*
      the result of this parse provides freshly-allocated strings
      to the caller.  presumably, the caller will want to use them,
      without yet another allocation with strdup().  if not, the
      caller should use xpin_ns_name_clear() to free the results of a
      prior parse, before calling a new parse, to avoid a memory
      leak.  here we just reset the pointers to NULL:
    */

    N->local_name = NULL;
    N->ns_uri = NULL;
    N->ns_prefix = NULL;
    N->qname = NULL;

    if(!name) return 0;

/*
    parse out up to 3 substrings from namespace triplet:

    ix = 0:  local_name \0
    ix = 1:  ns_uri | local_name \0
    ix = 2:  ns_uri | local_name | ns_prefix \0

    example shows '|' as separator (extra spaces for legibility)
*/

    /* intitalize substring vector to NULL: */
    for(ix = 0; ix < 3; ++ix){
        ss[ix] = NULL;
    }

    ix = 0;
    s0 = (char *)name;
    /* read up to 2 sep-delimited substrings: */
    while(((sN = strchr(s0, ns_sep)) != NULL) && (ix < 2)){
        sslen = sN - s0;
        if(sslen > 0){
            tmp = (char *)malloc(sslen + 1);
            if(!tmp){
                /* XXX alloc failure! */
                return -1;
            }
            memcpy(tmp, s0, sslen);
            tmp[sslen] = '\0';
            ss[ix] = tmp;
        }
        else{
            /* XXX unexpected empty namespace substring! */
            ss[ix] = NULL;
        }
        s0 = sN + 1;
        ++ix;
    }

    /* read the last (or only) substring: */
    sslen = strlen(s0);
    if(sslen > 0){
        tmp = (char *)malloc(sslen + 1);
        if(!tmp){
            /* XXX alloc failure! */
            return -1;
        }
        memcpy(tmp, s0, sslen);
        tmp[sslen] = '\0';
        ss[ix] = tmp;
    }
    else{
        /* XXX unexpected empty namespace substring! */
        ss[ix] = NULL;
    }

    /* now sort out which substrings are which: */
    if(ix == 0){
        /* no namespace: */
        N->local_name = ss[0];
        N->ns_uri = NULL;
        N->ns_prefix = NULL;
        N->qname = strdup(N->local_name);
        if(! N->qname){
            /* XXX alloc failure! */
                xpin_ns_name_clear(N);
                return -1;
        }
    }
    else if(ix == 1){
        /* namespace, but no prefix: */
        N->local_name = ss[1];
        N->ns_uri = ss[0];
        N->ns_prefix = NULL;
        N->qname = strdup(N->local_name);
        if(! N->qname){
            /* XXX alloc failure! */
                xpin_ns_name_clear(N);
                return -1;
        }
    }
    else if(ix == 2){
        /* namespace with prefix: */
        char  *qname = NULL;

        N->local_name = ss[1];
        N->ns_uri = ss[0];
        N->ns_prefix = ss[2];

        if((ss[2] == NULL) || (ss[1] == NULL)){
            /* XXX unexpected empty prefix or local_name! */
                xpin_ns_name_clear(N);
                return -1;
        }

        qname = (char *)malloc(strlen(ss[1]) + strlen(ss[2]) + 2);
        if(! qname){
            /* XXX alloc failure! */
                xpin_ns_name_clear(N);
                return -1;
        }
        strcpy(qname, ss[2]);
        strcat(qname, ":");
        strcat(qname, ss[1]);
        N->qname = qname;
    }
    else{
        /* not here! */
        assert(1);
    }

    return 0;
}/*end xpin_expat_ns3_parse()*/




static
void
xpin_ns_name_clear(struct xpin_ns_name *N)
{
    if(!N)  return;

    if(N->local_name) free(N->local_name);
    if(N->ns_uri)     free(N->ns_uri);
    if(N->ns_prefix)  free(N->ns_prefix);
    if(N->qname)      free(N->qname);
    N->local_name = NULL;
    N->ns_uri = NULL;
    N->ns_prefix = NULL;
    N->qname = NULL;

    return;
}/* end xpin_ns_name_clear()*/


static
void
xpin_expat_start_element_handler
(
    struct xpindoc    *X,
    const Xpin_Char   *name,
    const Xpin_Char  **atts
)
{
    struct xpin           *xpin = X->xpin;
    struct xpin_ns_name    ns = {NULL, NULL, NULL, NULL};
    char                   ns_sep = xpin->ns_sep;
    const Xpin_Char       *local_name = NULL;
    const Xpin_Char       *ns_uri = NULL;
    const Xpin_Char       *ns_prefix = NULL;
    const Xpin_Char       *qname = NULL;
    int                    attr_count = 0;
    struct xpin_attr     **attributes = NULL;
    int                    i = 0;

    if(xpin->ns_parser != 0){
        xpin_expat_ns3_parse(&ns, name, ns_sep);
        local_name = (const char *)ns.local_name;
        ns_uri = (const char *)ns.ns_uri;
        ns_prefix = (const char *)ns.ns_prefix;
        qname = (const char *)ns.qname;
    }
    else{
        local_name = (const char *)strdup(name);
        ns_uri = NULL;
        ns_prefix = NULL;
        qname = NULL;
    }
    if(!local_name){
        /* XXX raise error, alloc failure! */
    }

    /* attributes: convert to xpindoc format: */

    /* get the attributes: */
    attr_count = 0;
    while(atts[attr_count]) ++attr_count;
    attr_count /= 2;

    if(attr_count){
        attributes = (struct xpin_attr **)
            malloc(sizeof(struct xpin_attr *) * attr_count);
        if(!attributes){
            /* XXX handle error */
        }
        for(i = 0; i < attr_count; ++i){
            attributes[i] = (struct xpin_attr *)
                malloc(sizeof(struct xpin_attr));
            if(!attributes[i]){
                /* XXX handle error */
            }
            if(xpin->ns_parser != 0){
                /* namespace support: */
                xpin_expat_ns3_parse(&ns, atts[i*2], ns_sep);
                attributes[i]->name = (const char *)ns.qname;
                attributes[i]->local_name = (const char *)ns.local_name;
                attributes[i]->ns_uri = (const char *)ns.ns_uri;
                attributes[i]->ns_prefix = (const char *)ns.ns_prefix;
                attributes[i]->value = (const char *)strdup(atts[(i*2) + 1]);
                /* XXX error check strdup()! */
            }
            else{
                attributes[i]->name = (const char *)strdup(atts[i*2]);
                /* XXX error check strdup()! */
                attributes[i]->local_name = NULL;
                attributes[i]->ns_uri = NULL;
                attributes[i]->ns_prefix = NULL;
                attributes[i]->value = (const char *)strdup(atts[(i*2) + 1]);
                /* XXX error check strdup()! */
            }
            if(!attributes[i]->name || !attributes[i]->value){
                /* XXX handle error */
            }
        }
    }

    xpin_meta_start_element_handler
        (X, local_name, ns_uri, ns_prefix, qname, attr_count, attributes);

    return;
}/*end xpin_expat_start_element_handler()*/


static
void
xpin_expat_end_element_handler
(
    struct xpindoc    *X,
    const Xpin_Char   *name
)
{
/*
  xpin_meta_end_element doesn't need/use names again, because
  they are already installed in the node object.
*/
  
    xpin_meta_end_element_handler(X);

    return;
}/*end xpin_expat_end_element_handler()*/



/* xpin object methods (wrappers for expat parser): */
static
int
xpin_get_lineno(struct xpin *xpin)
{
    return XML_GetCurrentLineNumber(xpin->parser);
}


static
int
xpin_get_column(struct xpin *xpin)
{
    return XML_GetCurrentColumnNumber(xpin->parser);
}


static
int
xpin_get_size(struct xpin *xpin)
{
    return XML_GetCurrentByteCount(xpin->parser);
}


static
long int
xpin_get_offset(struct xpin *xpin)
{
    return XML_GetCurrentByteIndex(xpin->parser);
}



/* xpindoc object methods: */
static
int
xpindoc_ns_parser(struct xpindoc *X)
{
    return X->xpin->ns_parser;
}


static
void
xpindoc_set_mydata(struct xpindoc *X, void *mydata)
{
    X->xpin->mydata = mydata;
    return;
}


static
void *
xpindoc_get_mydata(struct xpindoc *X)
{
    return X->xpin->mydata;
}



static
int
xpindoc_set_datamode(struct xpindoc *X, int datamode)
{
    /* XXX error check datamode for legal value! */
    X->xpin->datamode = datamode;
    return 0;
}



static
int
xpindoc_get_datamode(struct xpindoc *X)
{
    return X->xpin->datamode;
}



static
void
xpindoc_ignore_on(struct xpindoc *X)
{
    X->xpin->ignore = XPIN_IGNORE_ON;
}



static
void
xpindoc_ignore_off(struct xpindoc *X)
{
    X->xpin->ignore = XPIN_IGNORE_OFF;
}



static
int
xpindoc_is_ignored(struct xpindoc *X)
{
    return X->xpin->ignore;
}



static
int
xpindoc_expat_parse_stream
(
    struct xpindoc *X,
    FILE           *stream
)
{
    struct xpin      *xpin = X->xpin;
    XML_Parser        parser = xpin->parser;
    void             *buffer;
    int               len = 0, done = 0;

    /* install xpindoc as data payload for expat: */
    XML_SetUserData(parser, X);

    /* install expat handlers with xpindoc meta handlers: */
    /*
    ** note our handler functions are cast to expat callback signatures
    **   -- this allows us to use xpindoc objects directly,
    **      instead of casting back from void pointers
    */
    XML_SetStartElementHandler(parser,
        (XML_StartElementHandler)&xpin_expat_start_element_handler);
    XML_SetEndElementHandler(parser,
        (XML_EndElementHandler)&xpin_expat_end_element_handler);

    XML_SetProcessingInstructionHandler(parser,
        (XML_ProcessingInstructionHandler)&xpin_meta_pi_handler);

    XML_SetStartCdataSectionHandler(parser,
        (XML_StartCdataSectionHandler)&xpin_meta_start_cdata_handler);
    XML_SetEndCdataSectionHandler(parser,
        (XML_EndCdataSectionHandler)&xpin_meta_end_cdata_handler);

    XML_SetCharacterDataHandler(parser,
        (XML_CharacterDataHandler)&xpin_meta_chardata_handler);


    /* call xpindoc start_document pseudo-handler: */
    xpin_meta_start_document_handler(X);

    while(1){
        /* halt support: */
        if(X->is_croaked(X) || X->is_halted(X))
            /* XXX return value other than failure? */
            return -1;

        /*
        **  NOTE:
        **
        **  xpin_raise_error() sets is_croaked TRUE
        */

        buffer = XML_GetBuffer(parser, PARSE_BUFFER_SIZE);
        if(buffer == NULL){
            xpin_raise_error(X, XPIN_ERROR_ALLOC,
                             "buffer allocation failure in parser");
            return XPIN_ERROR_ALLOC;
        }

        len = fread(buffer, 1, PARSE_BUFFER_SIZE, stream);
        if(ferror(stream)){
            xpin_raise_error(X, XPIN_ERROR_FILE,
                             "file read error during parse");
            return XPIN_ERROR_FILE;
        }

        done = feof(stream);
        if((XML_ParseBuffer(parser, len, done)) == 0){
            char  msgbuf[MSGBUF_SIZE];

            snprintf(msgbuf, MSGBUF_SIZE, "xml syntax error: %s",
                     XML_ErrorString(XML_GetErrorCode(parser)) );

            xpin_raise_error(X, XPIN_ERROR_PARSE, msgbuf);
            return XPIN_ERROR_PARSE;
        }

        if(done) break;
    }


    /* call xpindoc end_document pseudo-handler: */
    xpin_meta_end_document_handler(X);

    return 0;
}/*end xpindoc_expat_parse_stream()*/


static
void
xpindoc_halt_parse(struct xpindoc *X)
{
   X->xpin->is_halted = 1;
}/*end xpindoc_parse_halt()*/ 



static
int
xpindoc_is_halted(struct xpindoc *X)
{
    return X->xpin->is_halted;
}/*end xpindoc_is_halted()*/



static
int
xpindoc_is_croaked(struct xpindoc *X)
{
    return X->xpin->is_croaked;
}/*end xpindoc_is_croaked()*/


static
int
xpindoc_set_filter
(
    struct xpindoc *X,
    void  (*filter)(struct xpindoc *X, const char *chardata, int len)
)
{

    X->xpin->filter = filter;
    return 0;
}/*end xpindoc_set_filter()*/



static
struct xpin_event *
xpindoc_get_event(struct xpindoc *X)
{
    return X->xpin->event;
}


static
enum xpin_event_type
xpindoc_get_event_type(struct xpindoc *X)
{
    //return (X->event(X))->type(X->event(X));
    return X->xpin->event->type(X->xpin->event);
}


static
struct xpin_node *
xpindoc_get_node(struct xpindoc *X)
{
    return X->xpin->node;
}


static
struct xpin_pi *
xpindoc_get_pi(struct xpindoc *X)
{
    return X->xpin->pi;
}


static
struct xpin_error *
xpindoc_get_error(struct xpindoc *X)
{
    return X->xpin->error;
}


static
int
xpindoc_get_node_depth(struct xpindoc *X)
{
    return X->xpin->node->depth(X->xpin->node);
}


static
const char *
xpindoc_get_node_tagname(struct xpindoc *X)
{
    return X->xpin->node->tagname(X->xpin->node);
}


static
const char *
xpindoc_get_node_tagpath(struct xpindoc *X)
{
    return X->xpin->node->tagpath(X->xpin->node);
}


static
const char *
xpindoc_get_node_content(struct xpindoc *X)
{
    return X->xpin->node->content(X->xpin->node);
}


static
struct xpin_text *
xpindoc_get_node_text(struct xpindoc *X)
{
    return X->xpin->node->text(X->xpin->node);
}


static
int
xpindoc_push_output(struct xpindoc *X, void *dest, int dest_type)
{
    struct xpin_writer  *W = X->xpin->writer;

    return W->push_output(W, dest, dest_type);
}


static
void *
xpindoc_pop_output(struct xpindoc *X)
{
    struct xpin_writer  *W = X->xpin->writer;

    return W->pop_output(W);
}


static
int
xpindoc_output_char(struct xpindoc *X, char c)
{
    struct xpin_writer  *W = X->xpin->writer;

    return W->put_char(W, c);
}



static
int
xpindoc_output_string(struct xpindoc *X, const char *string)
{
    struct xpin_writer  *W = X->xpin->writer;

    return W->put_string(W, string);
}



static
int
xpindoc_output_chardata(struct xpindoc *X, const char *data, int len)
{
    struct xpin_writer  *W = X->xpin->writer;

    return W->put_chardata(W, data, len);
}

 
static
int
xpindoc_output_printf(struct xpindoc *X, const char *format, ...)
{
    va_list              ap;
    struct xpin_writer  *W = X->xpin->writer;
    char                *s = NULL;
    int                  err = 0;

    va_start(ap, format);
    err = vasprintf(&s, format, ap);
    va_end(ap);
    if((err < 0) || (s == NULL)){
        /* XXX error in vasprintf()! */
        if(s) free(s);
        return err;
    }

    err = W->put_string(W, (const char *)s);
    if(s) free(s);

    return err;
}
 
 

/* XpinDoc creator (exported): */
struct xpindoc *
new_xpindoc(int flags)
{
    struct xpindoc   *X = NULL;
    struct xpin      *xpin = NULL;

    X = (struct xpindoc *)malloc(sizeof(struct xpindoc));
    if(X == NULL)  return NULL;

    xpin = (struct xpin *)malloc(sizeof(struct xpin));
    if(xpin == NULL){
        free(X);
        return NULL;
    }

    X->xpin = xpin;

    if(flags & XPIN_NAMESPACE){
        /* use namespaces: */
        xpin->parser = XML_ParserCreateNS(NULL, XPIN_NSSEP);
        if(xpin->parser == NULL){
            free_xpindoc(X);
            return NULL;
        }
        XML_SetReturnNSTriplet(xpin->parser, 1);
        xpin->ns_parser = 1;
        xpin->ns_sep = XPIN_NSSEP;
        if(flags & XPIN_QNAME){
            xpin->using_qname = 1;
        }
        else{
            xpin->using_qname = 0;
        }
    }
    else{
        /* no namespaces: */
        xpin->parser = XML_ParserCreate(NULL);
        if(xpin->parser == NULL){
            free_xpindoc(X);
            return NULL;
        }
        xpin->ns_parser = 0;
        xpin->ns_sep = '\0';
        xpin->using_qname = 0;
    }


    xpin->error = new_xpin_error();
    if(xpin->error == NULL){
        free_xpindoc(X);
        return NULL;
    }

    xpin->writer = new_xpin_writer(xpin);
    if(xpin->writer == NULL){
        free_xpindoc(X);
        return NULL;
    }

    /*
    ** initialize an event object:
    ** (we will re-use this event object for subsequent events)
    */
    xpin->event = new_xpin_event();

    if(xpin->event == NULL){
        free_xpindoc(X);
        return NULL;
    }

    /*
    ** initialize a PI object:
    ** (we will re-use this PI object for subsequent PIs)
    */
    xpin->pi = new_xpin_pi();
    if(xpin->pi == NULL){
        free_xpindoc(X);
        return NULL;
    }

    xpin->suite = new_xpin_handler_suite();
    if(xpin->suite == NULL){
        free_xpindoc(X);
        return NULL;
    }

    /* initialize output_stack: */
    xpin->writer->push_output(xpin->writer, stdout, XPINOUT_FILE);

    xpin->node = NULL;

    xpin->datamode = XPIN_DEFAULT_DATAMODE;
    xpin->ignore = XPIN_IGNORE_OFF;
    xpin->is_halted = 0;
    xpin->is_croaked = 0;
    xpin->mydata = NULL;
    xpin->filter = NULL;

    xpin->line    = &xpin_get_lineno;
    xpin->column  = &xpin_get_column;
    xpin->size    = &xpin_get_size;
    xpin->offset  = &xpin_get_offset;

    X->ns_parser     = &xpindoc_ns_parser;

    X->set_mydata    = &xpindoc_set_mydata;
    X->mydata        = &xpindoc_get_mydata;

    X->set_datamode  = &xpindoc_set_datamode;
    X->datamode      = &xpindoc_get_datamode;

    X->ignore_on     = &xpindoc_ignore_on;
    X->ignore_off    = &xpindoc_ignore_off;
    X->is_ignored    = &xpindoc_is_ignored;

    X->parse_stream  = &xpindoc_expat_parse_stream;

    X->halt          = &xpindoc_halt_parse;
    X->is_halted     = &xpindoc_is_halted;

    X->is_croaked    = &xpindoc_is_croaked;

    X->set_handler   = &xpin_set_handler;

    X->set_filter    = &xpindoc_set_filter;

    X->event         = &xpindoc_get_event;
    X->event_type    = &xpindoc_get_event_type;

    X->node          = &xpindoc_get_node;

    X->pi            = &xpindoc_get_pi;

    X->error         = &xpindoc_get_error;

    X->depth         = &xpindoc_get_node_depth;
    X->tagname       = &xpindoc_get_node_tagname;
    X->tagpath       = &xpindoc_get_node_tagpath;
    X->content       = &xpindoc_get_node_content;
    X->node_text     = &xpindoc_get_node_text;

    /* access Writer + output_stack: */
    X->push_output     = &xpindoc_push_output;
    X->pop_output      = &xpindoc_pop_output;
    X->put_char        = &xpindoc_output_char;
    X->put_string      = &xpindoc_output_string;
    X->put_chardata    = &xpindoc_output_chardata;
    X->printf          = &xpindoc_output_printf;

    X->free            = &free_xpindoc;

    return X;
}/*end new_xpindoc()*/



static
void
free_xpindoc(struct xpindoc *X)
{
    struct xpin       *xpin = X->xpin;

    if(xpin != NULL){
        struct xpin_node  *node = xpin->node;

        /* free any remaining nodes back to root: */
        while(node != NULL){
            struct xpin_node  *parent = node->parent(node);
            free_xpin_node(node);
            node = parent;
        }

        XML_ParserFree(xpin->parser);
        if(xpin->event) free_xpin_event(xpin->event);
        if(xpin->pi) free_xpin_pi(xpin->pi);
        free_xpin_error(xpin->error);
        free_xpin_writer(xpin->writer);
        free_xpin_handler_suite(xpin->suite);
        free(xpin);
    }
    free(X);

    return;
}/*end free_xpindoc()*/



/* that's all, folks! */
