/* xpin_error.c
**
** XpinDoc
** an XML parse instance document processing utility
**
** Copyright (C) 2002, 2003, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.09.29
** ===
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "xpindoc.h"
#include "xpin.h"
#include "xpin_handlers.h"
#include "xpin_error.h"


#define  ERRMSG_BUFSIZE  256

/* opaque object internals: */
struct xpin_error_ {
    int     line;
    int     column;
    int     error_code;
    char    error_message[ERRMSG_BUFSIZE];
};

/* method prototypes: */
static
int
xpin_error_get_line(struct xpin_error *);

static
int
xpin_error_get_column(struct xpin_error *);

static
int
xpin_error_get_code(struct xpin_error *);

static
const char *
xpin_error_get_message(struct xpin_error *);


/* internal error dispatch function: */
static
void
xpin_error_dispatch(struct xpindoc *);

static
void
xpin_error_default_handler(struct xpindoc *);


/* CREATOR: */
struct xpin_error *
new_xpin_error(void)
{
    struct xpin_error   *E = NULL;
    struct xpin_error_  *error = NULL;

    E = (struct xpin_error *)malloc(sizeof(struct xpin_error));
    if(E == NULL)  return NULL;

    error = (struct xpin_error_ *)malloc(sizeof(struct xpin_error_));
    if(error == NULL){
        free(E);
        return NULL;
    }

    error->line = 0;
    error->column = 0;
    error->error_code = 0;
    error->error_message[0] = '\0';

    E->error = error;

    /* methods: */
    E->line    = &xpin_error_get_line;
    E->column  = &xpin_error_get_column;
    E->code    = &xpin_error_get_code;
    E->message = &xpin_error_get_message;

    return E;
}/*end new_xpin_error()*/


/* DESTRUCTOR: */
void
free_xpin_error(struct xpin_error *E)
{
    if(E){
        if(E->error) free(E->error);
        free(E);
    }

    return;
}/*end free_xpin_error()*/


/* methods: */
static
int
xpin_error_get_line(struct xpin_error *E)
{
    return E->error->line;
}/*end xpin_error_get_line()*/


static
int
xpin_error_get_column(struct xpin_error *E)
{
    return E->error->column;
}/*end xpin_error_get_column()*/


static
int
xpin_error_get_code(struct xpin_error *E)
{
    return E->error->error_code;
}/*end xpin_error_get_code()*/


static
const char *
xpin_error_get_message(struct xpin_error *E)
{
    return E->error->error_message;
}/*end xpin_error_get_message()*/



/* exported error system calls: */

void
xpin_raise_error
(
    struct xpindoc *X,
    int             code,
    const char     *message
)
{
    struct xpin        *xpin = X->xpin;
    struct xpin_error  *E = X->error(X);

    /* set line/column from parser: */
    E->error->line = xpin->line(xpin);
    E->error->column = xpin->column(xpin);

    E->error->error_code = code;

#if 0
    strncpy(E->error->error_message,
            message,
            ERRMSG_BUFSIZE - 1);
    E->error->error_message[ERRMSG_BUFSIZE] = '\0';
#endif

    /* use strlcpy(): */
    strlcpy(E->error->error_message,
            message, sizeof(E->error->error_message) );


    xpin_error_dispatch(X);
    /* croak parse if return from dispatch: */
    xpin->is_croaked = 1;
    return;
}/*end xpin_raise_error()*/




/* error system internals: */

static
void
xpin_error_dispatch(struct xpindoc *X)
{
    if(X->xpin->suite->error_handler){
        /* call user-installed error handler: */
        X->xpin->suite->error_handler(X);
    }
    else{
        xpin_error_default_handler(X);
    }
    
}/*end xpin_error_dispatch()*/


static
void
xpin_error_default_handler(struct xpindoc *X)
{
    struct xpin_error  *E = X->error(X);

    fprintf(stderr, "XpinDoc error! (default error handler):\n");
    fprintf(stderr, "  input line %d [column %d]: %s\n",
            E->line(E), E->column(E), E->message(E));
    fprintf(stderr, "  active tagpath: %s\n", X->tagpath(X));
    fprintf(stderr, "  active tagname: %s\n", X->tagname(X));
    return;
}/*end xpin_error_default_handler()*/



/* that's all, folks! */
