/* xpin_event.c
**
** XpinDoc
** an XML parse instance document processing utility
**
** Copyright (C) 2002, 2003, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.09.22
** ===
*/
#include <stdlib.h>

/* our headers: */
#include "xpindoc.h"
#include "xpin_event.h"


/* EVENT OBJECTS */

/* internals: */
struct xpin_event_ {
    /* type of this event (see enums): */
    enum xpin_event_type    type;
    /* line number of event (unit offset): */
    int                     line;
    /* column in line (zero offset): */
    int                     column;
    /* length of event (incl null!): */
    int                     size;
    /* byte offset in stream: */
    long int                index;
    /* data can be xpin_node, xpin_pi, comment: */
    void                   *data;
};    



/* internal prototypes: */

/* EVENT METHODS: */
inline
static
enum xpin_event_type
event_get_type(struct xpin_event *);

inline
static
int
event_get_line(struct xpin_event *);

inline
static
int
event_get_column(struct xpin_event *);

inline
static
int
event_get_size(struct xpin_event *);

inline
static
long int
event_get_index(struct xpin_event *);

inline
static
void *
event_get_data(struct xpin_event *);

inline
static
struct xpin_pi *
event_get_pi(struct xpin_event *);


/* event method definitions: */
static
enum xpin_event_type
event_get_type(struct xpin_event *E)
{
    return E->event->type;
}

static
int
event_get_line(struct xpin_event *E)
{
    return E->event->line;
}

static
int
event_get_column(struct xpin_event *E)
{
    return E->event->column;
}

static
int
event_get_size(struct xpin_event *E)
{
    return E->event->size;
}

static
long int
event_get_index(struct xpin_event *E)
{
    return E->event->index;
}

static
void *
event_get_data(struct xpin_event *E)
{
    return E->event->data;
}


static
struct xpin_pi *
event_get_pi(struct xpin_event *E)
{
    return (struct xpin_pi *)E->event->data;
}


/* exported (xpin_event.h): */

/* EVENT CONSTRUCTOR: */
struct xpin_event *
new_xpin_event()
{
    struct xpin_event    *E = NULL;
    struct xpin_event_   *event = NULL;

    E = (struct xpin_event *)malloc(sizeof(struct xpin_event));
    if(E == NULL)  return NULL;

    event = (struct xpin_event_ *)malloc(sizeof(struct xpin_event_));
    if(event == NULL){
        free(E);
        return NULL;
    }

    event->type   = XPIN_NULL;
    event->line   = 0;
    event->column = 0;
    event->size   = 0;
    event->index  = 0L;
    event->data   = (void *)NULL;

    E->event = event;

    /* methods: */
    E->type    = &event_get_type;
    E->line    = &event_get_line;
    E->column  = &event_get_column;
    E->size    = &event_get_size;
    E->index   = &event_get_index;
    E->data    = &event_get_data;
    E->pi      = &event_get_pi;

    return E;
}/*end new_xpin_event()*/



/* RE-USE EVENT OBJECT: */
struct xpin_event *
set_xpin_event
(
    struct xpin_event     *E,
    enum xpin_event_type   type,
    int                    line,
    int                    column,
    int                    size,
    long int               index,
    void                  *data
)
{
    if(!E || !E->event) return NULL;

    E->event->type   = type;
    E->event->line   = line;
    E->event->column = column;
    E->event->size   = size;
    E->event->index  = index;
    E->event->data   = data;

    return E;
}/*end set_xpin_event()*/



/* EVENT DESTRUCTOR: */
void
free_xpin_event(struct xpin_event *E)
{
   if(E == NULL)
       return;

   /* XXX freeing E->event->data? */

   /* event->data is variously:

          start_document:  NULL
          end_document:    NULL
          start_element:   node
          end_element:     NULL
          pi:              pi

      node and pi are in xpindoc object, and
      should not be deallocated here
    */

   if(E->event) free(E->event);
   free(E);

   return;
}/*end free_xpin_event()*/ 



/* that's all, folks! */
