/* xpin_handlers.c
**
** XpinDoc
** an XML parse instance document processing utility
**
** Copyright (C) 2002, 2003, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.10.01
** ===
*/
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fnmatch.h>


/* our headers: */
#include "dt_hash.h"
#include "dt_list.h"
#include "xpindoc.h"
#include "xpin_event.h"
#include "xpin_node.h"
#include "xpin_pi.h"
#include "xpin_text.h"
#include "xpin_handlers.h"
#include "xpin.h"


/* pick a prime: */
#define  HASHTBL_SIZE  499


/* forms of keystr to xpin_set_handler(): */
enum keystr_type {
    KEY_ERROR = -1,
    KEY_NULL =   0,
    KEY_KEYWORD,
    KEY_TAGNAME,
    KEY_PI,
    KEY_TAGPATH
};


/* HANDLER OBJECTS: */

struct xpin_handler_item {
    /*
    ** pattern is extracted from keystr argument to function:
    **
    **   XPIN->set_handler(XPIN, keystr, handler);
    **
    ** where keystr is one of:
    **
    **   KEY_TAGNAME:  tagname expression (eg: "<tag>" or "</tag>")
    **   KEY_PI:       pi target expression (eg: "<?pi?>")
    **   KEY_TAGPATH:  tagpath expression (eg: "/tagpath")
    **   KEY_KEYWORD:  default handler keyword (eg: "start_handler")
    **
    */
    const char            *pattern;
    Xpin_Handler           handler;
};



/* prototypes for functions internal to this module: */

static
struct xpin_handler_item *
new_xpin_handler_item
(
    const char            *pattern,
    Xpin_Handler           handler
);

static
void
free_xpin_handler_item(struct xpin_handler_item *);

static
int
xpin_keystr_type(const char *keystr);

static
int
xpin_set_tagname_handler
(
    struct xpindoc    *X,
    const char        *keystr,
    Xpin_Handler       handler
);

static
int
xpin_set_tagpath_handler
(
    struct xpindoc    *X,
    const char        *keystr,
    Xpin_Handler       handler
);

static
int
xpin_set_pi_handler
(
    struct xpindoc    *X,
    const char        *keystr,
    Xpin_Handler       handler
);

static
int
xpin_keycmp(const char *k1, const char *k2);

static
int
xpin_set_keyword_handler
(
    struct xpindoc    *X,
    const char        *keystr,
    Xpin_Handler       handler
);


static
int
xpin_handler_dispatch(struct xpindoc *);


/* hash function parameters: */
static
unsigned int
xpin_hasher(const void *key);

inline
static
int
xpin_hash_compare(const void *key1, const void *key2);


/* dt_list tagpath match function: */
static
int
xpin_tagpath_match(const void *item, const void *tagpath);




/* definitions: */
static
struct xpin_handler_item *
new_xpin_handler_item
(
    const char           *pattern,
    Xpin_Handler          handler
)
{
    struct xpin_handler_item  *hitem = NULL;

    hitem = (struct xpin_handler_item *)malloc(sizeof(struct xpin_handler_item));
    if(hitem == NULL)
        return NULL;

    hitem->pattern = pattern;
    hitem->handler = handler;

    return hitem;
}/*end new_xpin_handler_item()*/



static
void
free_xpin_handler_item(struct xpin_handler_item *hitem)
{
    if(hitem){
        if(hitem->pattern) free((char *)hitem->pattern);
        free(hitem);
    }

    return;
}/*end free_xpin_handler_item()*/



/*
**    "<tag>"  start element
**    "</tag>" end element
**    "<tag/>" start & end element
**
**    "S:/tagpath"  start element
**    "E:/tagpath"  end element
**    "A:/tagpath"  start & end element
**
**    "<?pi?>  processing instruction
**
**    "keyword"
*/

static
int
xpin_keystr_type(const char *keystr)
{
    int  keylen = 0;

    if(keystr == NULL)
        return KEY_NULL;
   
    keylen = strlen(keystr);
    if(keylen < 3)
        return KEY_ERROR;

    if(keystr[0] == '<')
    {
        if(keystr[keylen - 1] != '>')
            return KEY_ERROR;

        if((keystr[1] == '/')  && ((keylen < 4) || (keystr[keylen - 2] == '/')) )
            return KEY_ERROR;

        if((keystr[keylen - 2] == '/')  && (keylen < 4))
            return KEY_ERROR;

        if(keystr[1] == '?'){
            if((keystr[keylen - 2] != '?') && (keylen < 5)) {
                return KEY_ERROR;
            }
            else{
                return KEY_PI;
            }
        }

        /* we assume that leaves: */
        return KEY_TAGNAME;
    }

    else if(keystr[0] == '/')
    {
        return KEY_TAGPATH;
    }

    /* XXX assume keyword?? */
    else{
        return KEY_KEYWORD;
    }

    /* can't be here: */
    return KEY_ERROR;

}/*end xpin_keystr_type()*/


static
char *
xpin_extract_key_pattern
(
    const char  *keystr,
    int          patt_offset,
    int          patt_length
);


static
char *
xpin_extract_key_pattern
(
    const char  *keystr,
    int          patt_offset,
    int          patt_length
)
{
    char        *pattern = NULL, *pp = NULL;
    const char  *kp = NULL;
    int          i = 0;

    if(!keystr) return NULL;

    /* extract pattern from keystr: */
    kp = keystr + patt_offset;
    pattern = (char *)malloc(patt_length + 1);
    if(pattern == NULL)
        /* XXX error! */
        return NULL;

    pp = pattern;
    for(i = 0; i < patt_length; ++i){
        *pp++ = *kp++;
    }
    *pp = '\0';

    return pattern;
}/*xpin_extract_key_pattern()*/



/*
** for:
**
**     xpin_set_tagname_handler()
**     xpin_set_tagpath_handler()
**     xpin_set_keyword_handler()
**
** called only from xpin_set_handler(), and only after successful
** return from xpin_keystr_type()
*/

static
int
xpin_set_tagname_handler
(
    struct xpindoc    *X,
    const char        *keystr,
    Xpin_Handler       handler
)
{
    int                        keylen = 0;
    dt_Hash                    tag_hash0 = NULL, tag_hash1 = NULL;
    int                        patt_offset = 0, patt_length = 0;
    char                      *pattern = NULL;

    keylen = strlen(keystr);

    if(keystr[1] == '/')
    {
        /* end element handler: */
        patt_offset = 2;
        patt_length = keylen - 3;
        tag_hash1 = X->xpin->suite->end_tag_hash;
    }
    else if(keystr[keylen - 2] == '/')
    {
        /* start & end element handler: */
        patt_offset = 1;
        patt_length = keylen - 3;
        tag_hash0 = X->xpin->suite->start_tag_hash; 
        tag_hash1 = X->xpin->suite->end_tag_hash;
    }
    else
    {
        /* start element handler: */
        patt_offset = 1;
        patt_length = keylen - 2;
        tag_hash0 = X->xpin->suite->start_tag_hash;
    }

    /* extract pattern from keystr: */
    /* pattern is allocated */
    pattern = xpin_extract_key_pattern(keystr, patt_offset, patt_length);
    if(!pattern){
        /* XXX error! */
        return -1;
    }

    /* duplicate key check: */
    if( (tag_hash0 && tag_hash0->exists(tag_hash0, pattern))
        || (tag_hash1 && tag_hash1->exists(tag_hash1, pattern)) )
    {
        /* XXX duplicate key error! */
        free(pattern);
        return -1;
    }


    /*
        Note:
        separate item allocations required for each hash table
    */

    /* XXX make this code block into a function! */

    /* add handler item to hash table(s): */
    /* add start handler: */
    if(tag_hash0){
        const char                *patt = NULL;
        struct xpin_handler_item  *hitem = NULL;

        patt = (const char *)strdup(pattern);
        if(!patt){
            /* XXX error! */
            free(pattern);
            return -1;
        }
        hitem = new_xpin_handler_item(patt, handler);
        if(hitem == NULL){
            /* XXX error! */
            free((char *)patt);
            free(pattern);
            return -1;
        }
        
        tag_hash0->insert(tag_hash0, patt, hitem);
    }

    /* add end handler: */
    if(tag_hash1){
        const char                *patt = NULL;
        struct xpin_handler_item  *hitem = NULL;

        patt = (const char *)strdup(pattern);
        if(!patt){
            /* XXX error! */
            free(pattern);
            return -1;
        }
        hitem = new_xpin_handler_item(patt, handler);
        if(hitem == NULL){
            /* XXX error! */
            free((char *)patt);
            free(pattern);
            return -1;
        }
        
        tag_hash1->insert(tag_hash1, patt, hitem);
    }


    /* XXX no error checking on hash insertions! */

    /* deallocate pattern used for duplicate key check: */
    free(pattern);
    return 0;        
}/*end xpin_set_tagname_handler()*/



static
int
xpin_set_tagpath_handler
(
    struct xpindoc    *X,
    const char        *keystr,
    Xpin_Handler       handler
)
{
    dt_List                     tagpath_list = X->xpin->suite->tagpath_list;
    const char                 *tagpath = NULL;
    struct xpin_handler_item   *hitem = NULL;

    tagpath = strdup(keystr);
    if(!tagpath){
        /* XXX error! */
        return -1;
    }

    hitem = new_xpin_handler_item(tagpath, handler); 
    if(!hitem){
        /* XXX error! */
        free((char *)tagpath);
        return -1;
    }

    tagpath_list->addend(tagpath_list, hitem);

    /* XXX no error checking on list insertions! */

    return 0;
}/*end xpin_set_tagpath_handler()*/


static
int
xpin_set_pi_handler
(
    struct xpindoc    *X,
    const char        *keystr,
    Xpin_Handler       handler
)
{
    int                        keylen = 0;
    dt_Hash                    pi_hash = X->xpin->suite->pi_hash;
    int                        patt_offset = 0, patt_length = 0;
    const char                *pi_name = NULL;
    struct xpin_handler_item  *hitem = NULL;

    keylen = strlen(keystr);

    patt_offset = 2;
    patt_length = keylen - 4;

    /* extract pi_name from keystr: */
    /* pi_name is allocated */
    pi_name = xpin_extract_key_pattern(keystr, patt_offset, patt_length);
    if(!pi_name){
        /* XXX error! */
        return -1;
    }

    /* duplicate key check: */
    if(pi_hash && pi_hash->exists(pi_hash, pi_name)) {
        /* XXX duplicate key error! */
        free((char *)pi_name);
        return -1;
    }

    hitem = new_xpin_handler_item(pi_name, handler);
    if(hitem == NULL){
        /* XXX error! */
        free((char *)pi_name);
        return -1;
    }

    /* XXX no error checking on hash insert! */
    pi_hash->insert(pi_hash, pi_name, hitem);

    return 0;
}/*end xpin_set_pi_handler()*/



static
int
xpin_keycmp(const char *k1, const char *k2)
{
    const unsigned char  *s1 = (const unsigned char *)k1,
                         *s2 = (const unsigned char *)k2;

    if(s1 == s2)
        return 0;

    /* case insensitive comparison: */
    while(tolower(*s1) == tolower(*s2)){
        if(*s1 == '\0'){
            return 0;
        }
        ++s1, ++s2;
    }

    return *s1 - *s2;

}/*end xpin_keycmp()*/



static
int
xpin_set_keyword_handler
(
    struct xpindoc    *X,
    const char        *keystr,
    Xpin_Handler       handler
)
{

    if(xpin_keycmp("start_document", keystr) == 0){
        X->xpin->suite->start_document_handler = handler;
        return 0;
    }

    else if(xpin_keycmp("end_document", keystr) == 0){
        X->xpin->suite->end_document_handler = handler;
        return 0;
    }

    else if(xpin_keycmp("start_element", keystr) == 0){
        X->xpin->suite->start_element_handler = handler;
        return 0;
    }

    else if(xpin_keycmp("end_element", keystr) == 0){
        X->xpin->suite->end_element_handler = handler;
        return 0;
    }

    else if(xpin_keycmp("ns_start_element", keystr) == 0){
        X->xpin->suite->ns_start_element_handler = handler;
        return 0;
    }

    else if(xpin_keycmp("ns_end_element", keystr) == 0){
        X->xpin->suite->ns_end_element_handler = handler;
        return 0;
    }

    else if(xpin_keycmp("processing_instruction", keystr) == 0){
        X->xpin->suite->pi_handler = handler;
        return 0;
    }

    else if(xpin_keycmp("start_cdata", keystr) == 0){
        X->xpin->suite->start_cdata_handler = handler;
        return 0;
    }

    else if(xpin_keycmp("end_cdata", keystr) == 0){
        X->xpin->suite->end_cdata_handler = handler;
        return 0;
    }

    else if(xpin_keycmp("error", keystr) == 0){
        X->xpin->suite->error_handler = handler;
        return 0;
    }

    /* ... more default handler keywords ... */

    /* if here, keystr is unknown: */
    return -1;
}/*end xpin_set_keyword_handler()*/



/*
** xpin_set_handler:
**   (this method is installed in XpinDoc object)
**
**   keystr in one of:
**
**     default handler keyword (eg, "start_handler")
**     nodename pattern (eg, "<tag>", or "</tag>")
**     tagpath pattern (eg, "/tag")
*/

int
xpin_set_handler
(
    struct xpindoc    *X,
    const char        *keystr,
    Xpin_Handler       handler
)
{
    enum keystr_type   keytype = 0;
    int                err = 0;

    if(!X || !keystr)  return -1;

    keytype = xpin_keystr_type(keystr);

    switch(keytype){
        case KEY_TAGNAME:
            err = xpin_set_tagname_handler(X, keystr, handler);
            break;

        case KEY_PI:
            err = xpin_set_pi_handler(X, keystr, handler);
            break;

        case KEY_TAGPATH:
            err = xpin_set_tagpath_handler(X, keystr, handler);
            break;

        case KEY_KEYWORD:
            err = xpin_set_keyword_handler(X, keystr, handler);
            break;

        default:
            err =  -1;
    }

    return err;
}/*end xpin_set_handler()*/



static
unsigned int
xpin_hasher(const void *key)
{
    unsigned int    h = 0;  
    unsigned char  *p = (unsigned char *)key;

    /* XXX hash function should barf on NULL key: */
    if(key == NULL){
        /* XXX handler error! */
    }

    /* note: okay--and expected--for h to wrap: */
    while(*p != '\0'){
        h = 37 * h + *p; 
        ++p;
    }

    /* return value will then be taken modulo tblsize: */
    return h;
};



static
int
xpin_hash_compare(const void *key1, const void *key2)
{
    return strcmp((const char *)key1, (const char *)key2);
}



/* list tagpath match function: */
static
int
xpin_tagpath_match(const void *item, const void *tagpath)
{
    /* XXX flags for fnmatch?? */
    return fnmatch(
        (const char *)((struct xpin_handler_item *)item)->pattern,
        (const char *)tagpath,
        0);
}/*end xpin_tagpath_match()*/



struct xpin_handler_suite *
new_xpin_handler_suite(void)
{
    struct xpin_handler_suite  *S = NULL;

    S = (struct xpin_handler_suite *)
            malloc(sizeof(struct xpin_handler_suite));
    if(S == NULL)  return NULL;

    /* initialize default handlers: */
    S->start_document_handler = NULL;
    S->end_document_handler = NULL;
    S->start_element_handler = NULL;
    S->end_element_handler = NULL;
    S->ns_start_element_handler = NULL;
    S->ns_end_element_handler = NULL;
    S->pi_handler = NULL;
    S->start_cdata_handler = NULL;
    S->end_cdata_handler = NULL;
    /* ... */

    /* allow safe deallocation on error: */
    S->start_tag_hash = NULL;
    S->end_tag_hash = NULL;
    S->pi_hash = NULL;
    S->tagpath_list = NULL;

    /* initialize handler hash tables: */
    S->start_tag_hash =
        new_hash(HASHTBL_SIZE, &xpin_hasher, &xpin_hash_compare);
    if(!S->start_tag_hash){
        free_xpin_handler_suite(S);
        return NULL;
    }

    S->end_tag_hash =
        new_hash(HASHTBL_SIZE, &xpin_hasher, &xpin_hash_compare);
    if(!S->end_tag_hash){
        free_xpin_handler_suite(S);
        return NULL;
    }

    S->pi_hash =
        new_hash(HASHTBL_SIZE, &xpin_hasher, &xpin_hash_compare);
    if(!S->pi_hash){
        free_xpin_handler_suite(S);
        return NULL;
    }
    /* ... */

    /* initialize handler lists: */
    S->tagpath_list = new_list();
    if(!S->tagpath_list){
        free_xpin_handler_suite(S);
        return NULL;
    }

    return S;
}/*end new_xpin_handler_suite()*/


static
void
free_xpin_hash_item(void *key, void *data);


/* deallocator used for hash->clear(): */
static
void
free_xpin_hash_item(void *key, void *data)
{
    struct xpin_handler_item  *hitem =
        (struct xpin_handler_item *)data;

    /* key points to pattern, pattern also in hitem,
       so just free the hitem
    */
    if(hitem){
        if(hitem->pattern)  free((char *)hitem->pattern);
        free(hitem);
    }

    return;
}/*end free_xpin_hash_item()*/

void
free_xpin_handler_suite(struct xpin_handler_suite *S)
{
    if(!S) return;

    if(S->start_tag_hash){
        dt_Hash  H = S->start_tag_hash;

        H->clear(H, &free_xpin_hash_item);
        H->free(H);
        H = NULL;
    }

    if(S->end_tag_hash){
        dt_Hash  H = S->end_tag_hash;

        H->clear(H, &free_xpin_hash_item);
        H->free(H);
        H = NULL;
    }

    if(S->pi_hash){
        dt_Hash  H = S->pi_hash;

        H->clear(H, &free_xpin_hash_item);
        H->free(H);
        H = NULL;
    }

    if(S->tagpath_list){
        dt_List  L = S->tagpath_list;
        struct xpin_handler_item  *hitem = NULL;

        while(L->items(L)){
            hitem = L->pop(L);
            free_xpin_handler_item(hitem);
        }
        L->free(L);
        L = NULL;
    }

    free(S);
    return;
}/*end free_xpin_handler_suite()*/



/*
** "pseudo" handlers: meta start_document and end_document
**    declared extern in "xpin_handlers.h"
**    called in xpindoc.c module before/after expat parse:
**    (expat does not generate calls for these events)
*/

void
xpin_meta_start_document_handler(struct xpindoc *X)
{
    struct xpin    *xpin = X->xpin;
    const char     *local_name = NULL, *qname = NULL;
    const char     *tagname = NULL, *tagpath = NULL;

    /* halt support: */
    if(X->is_halted(X))
        return;

    assert(xpin->event != NULL);

    /*
    ** initialize the start document event:
    */
    xpin->event = set_xpin_event(
        xpin->event,
        XPIN_START_DOCUMENT,  /*type*/
        0,                    /*line*/
        0,                    /*column*/
        0,                    /*size*/
        0L,                   /*index*/
        (void *)NULL          /*data for this event*/ );

    if(xpin->event == NULL){
        /* XXX handle error! */
    }

    /*  note:
        dynamic alloc of tagname/tagpath required for consistency in 
        freeing node chain back to root element in free_xpindoc()
    */
    if(xpin->using_qname){
        qname = (const char *)strdup("/");
        tagname = qname;
    }
    else{
        local_name = (const char *)strdup("/");
        tagname = local_name;
    }

    if(tagname == NULL){
        /* XXX handle error! */
    }

    tagpath = (const char *)strdup("/");
    if(tagpath == NULL){
        /* XXX handle error! */
    }

    /* initialize data for root node: */
    xpin->node = new_xpin_node(
        0,         /*depth of root element*/
        tagname,   /*name of root element*/
        tagpath,   /*tagpath*/
        local_name,
        NULL,      /*ns_uri*/
        NULL,      /*ns_prefix*/
        qname,
        0,         /*attr_count */
        NULL,      /*attributes */
        NULL       /*parent*/  );

    if(xpin->node == NULL){
        /* XXX handle error! */
    }

    /* dispatch to user's start_document handler: */
    xpin_handler_dispatch(X);
 
    return;
}/*end xpin_meta_start_document_handler()*/



void
xpin_meta_end_document_handler(struct xpindoc *X)
{
    struct xpin    *xpin = X->xpin;

    /* halt support: */
    if(X->is_halted(X))
        return;

    /* initialize this event: */
    xpin->event = set_xpin_event(
        xpin->event,
        XPIN_END_DOCUMENT,
        xpin->line(xpin),
        xpin->column(xpin),
        xpin->size(xpin),
        xpin->offset(xpin),
        (void *)NULL  /*data for this event*/ );

    /* dispatch to user's end_document handler: */
    xpin_handler_dispatch(X);

    /* XXX other cleanup at end of document?? */

    return;
}/*end xpin_meta_end_document_handler()*/



void
xpin_meta_start_element_handler
(
    struct xpindoc      *X,     /* our XpinDoc object */
    const Xpin_Char     *local_name, /* local_name */
    const Xpin_Char     *ns_uri,     /* namespace */
    const Xpin_Char     *ns_prefix,  /* namespace */
    const Xpin_Char     *qname,      /* qualified name */
    int                  attr_count, /* attr_count */
    struct xpin_attr   **attributes  /* attributes */
)
{
    struct xpin        *xpin = X->xpin;
    int                 depth = xpin->node->depth(xpin->node);
    const char         *tagname = NULL;
    const char         *tagpath = xpin->node->tagpath(xpin->node);
    const char         *newpath = NULL;
    struct xpin_node   *newnode = NULL;

    /* halt support: */
    if(X->is_halted(X))
        return;

    /* increment depth: */
    ++depth;

    if(xpin->using_qname){
        tagname = qname;
    }
    else {
        tagname = local_name;
    }

    if(!tagname){
        /* XXX handle error */
        assert(1);
    }

    /* construct tagpath to this element: */
    newpath = (char *)malloc(strlen(tagpath) + strlen(tagname) + 2);
    if(newpath == NULL){
        /* XXX handle error */
    }
    strcpy((char *)newpath, tagpath);
    if(depth > 1) strcat((char *)newpath, "/");
    strcat((char *)newpath, tagname);

    /* allocate and populate new node, adding to head of node list: */
    newnode = new_xpin_node(
        depth,
        tagname,
        newpath,
        local_name,
        ns_uri,
        ns_prefix,
        qname,
        attr_count,
        attributes,
        xpin->node   /* becomes parent */ );

    if(newnode == NULL){
        /* XXX handle error */
    }
    xpin->node = newnode;

    /* initialize this event: */
    xpin->event = set_xpin_event(
        xpin->event,
        XPIN_START_ELEMENT,
        xpin->line(xpin),
        xpin->column(xpin),
        xpin->size(xpin),
        xpin->offset(xpin),
        (void *)newnode );

    if(xpin->event == NULL){
        /* XXX handle error?? */
    }

    /* dispatch to user's handlers: */
    xpin_handler_dispatch(X);

    /* all done! */
    return;
}/*end xpin_meta_start_element_handler()*/



void
xpin_meta_end_element_handler
(
    struct xpindoc     *X
)
{
    struct xpin       *xpin = X->xpin;
    struct xpin_node  *node = xpin->node;

    /* halt support: */
    if(X->is_halted(X))
        return;

    /* initialize this event: */
    xpin->event = set_xpin_event(
        xpin->event,
        XPIN_END_ELEMENT,
        xpin->line(xpin),
        xpin->column(xpin),
        xpin->size(xpin),
        xpin->offset(xpin),
        (void *)node );

    if(xpin->event == NULL){
        /* XXX handle error?? */
    }

    /* dispatch to user's handler: */
    xpin_handler_dispatch(X);

    /* pop and free current node and replace with parent: */
    xpin->node = node->parent(node);
    free_xpin_node(node);

    return;
}/*end xpin_meta_end_element_handler()*/



void
xpin_meta_pi_handler
(
    struct xpindoc    *X,
    const Xpin_Char   *pi_name,
    const Xpin_Char   *pi_data
)
{
    struct xpin       *xpin = X->xpin;
    struct xpin_pi    *pi = xpin->pi;
    const char        *name = NULL, *data = NULL;

    /* halt support: */
    if(X->is_halted(X))
        return;

    /*
    ** our implementation stores the current PI in
    ** the XpinDoc object, until displaced by another
    ** PI.
    **
    ** this means XpinDoc applications can always
    ** access the current PI "in effect" for any
    ** event.
    */

    /* make copy of name, data: */
    name = strdup(pi_name);
    if(!name){
        /* XXX handle error */
    }

    /*
    ** data can be empty? if so, make a nul string:
    **
    **   note: expat does pass a nul string for empty data
    **   but we'll just make sure
    **
    **   also note:  empty name ("target") is an XML error
    */
    if(pi_data)
        data = strdup(pi_data);
    else
        data = strdup("\0");

    if(!data){
        /* XXX handle error */
    }

    /* free any previous pi: */
    pi = clear_xpin_pi(pi);

    /* re-use pi object: */
    pi = set_xpin_pi(pi, name, data);

    /* initialize this event: */
    xpin->event = set_xpin_event(
        xpin->event,
        XPIN_PI,
        xpin->line(xpin),
        xpin->column(xpin),
        xpin->size(xpin),
        xpin->offset(xpin),
        (void *)pi );

    if(xpin->event == NULL){
        /* XXX handle error?? */
    }

    /* dispatch to user's handler: */
    xpin_handler_dispatch(X);

    return;
}/*end xpin_meta_pi_handler()*/



void
xpin_meta_start_cdata_handler(struct xpindoc *X)
{
    struct xpin       *xpin = X->xpin;

    /* initialize this event: */
    xpin->event = set_xpin_event(
        xpin->event,
        XPIN_START_CDATA,
        xpin->line(xpin),
        xpin->column(xpin),
        xpin->size(xpin),
        xpin->offset(xpin),
        (void *)NULL );

    if(xpin->event == NULL){
        /* XXX handle error?? */
    }

    /* dispatch to user's handler: */
    xpin_handler_dispatch(X);

    return;
}/* end xpin_meta_start_cdata_handler()*/



void
xpin_meta_end_cdata_handler(struct xpindoc *X)
{
    struct xpin       *xpin = X->xpin;

    /* initialize this event: */
    xpin->event = set_xpin_event(
        xpin->event,
        XPIN_END_CDATA,
        xpin->line(xpin),
        xpin->column(xpin),
        xpin->size(xpin),
        xpin->offset(xpin),
        (void *)NULL );

    if(xpin->event == NULL){
        /* XXX handle error?? */
    }

    /* dispatch to user's handler: */
    xpin_handler_dispatch(X);

    return;
}/* end xpin_meta_end_cdata_handler()*/



/*
** generic content grabber:
*/
void
xpin_meta_chardata_handler
(
    struct xpindoc    *X,
    const Xpin_Char   *chardata,
    int                len
)
{
    /* halt || ignore support: */
    if(X->is_halted(X) || X->is_ignored(X))
        return;

    /*
    ** append chardata to content of current node or
    ** output to ouput_stack destination, through internal Writer
    ** object
    **
    ** note: incoming chardata is not nul terminated; these
    ** handlers do provide nul terminators
    */
    if(X->xpin->filter){
        X->xpin->filter(X, chardata, len);
    }
    else{
        X->put_chardata(X, chardata, len);
    }

    return; 
}/*end xpin_meta_chardata_handler()*/



static
int
xpin_handler_dispatch(struct xpindoc *X)
{
    int                         type = X->event_type(X);
    dt_Hash                     tag_hash = NULL;
    dt_List                     tagpath_list = NULL;
    struct xpin_handler_item   *hitem = NULL;

    /* halt support: we shouldn't get here if halted!*/
    if(X->is_halted(X)){
        fprintf(stderr, "logic error: halted in handler dispatch\n");
        return -1;
    }

    /* call handler based on event type: */        
    switch(type){
        case XPIN_START_DOCUMENT:
            {
                if(X->xpin->suite->start_document_handler)
                    X->xpin->suite->start_document_handler(X);
                break;
            }
        case XPIN_END_DOCUMENT:
            {
                if(X->xpin->suite->end_document_handler)
                    X->xpin->suite->end_document_handler(X);
                break;
            }
        case XPIN_START_ELEMENT:
            {
                /* if namespace,  ns_start_element handler: */
                struct xpin_node *N = X->node(X);

                if((X->xpin->ns_parser != 0)
                    && (N->ns_uri(N) !=  NULL)  
                    && (X->xpin->suite->ns_start_element_handler))
                {
                    X->xpin->suite->ns_start_element_handler(X);
                    break;
                }

                /* test for tagpath in start_tagpath_list: */
                tagpath_list = X->xpin->suite->tagpath_list;
                hitem = tagpath_list->search(
                          tagpath_list,
                          &xpin_tagpath_match,
                          X->tagpath(X));
                if(hitem){
                    hitem->handler(X);
                    break;
                }

                /* test for tag in start_tag_hash: */
                tag_hash = X->xpin->suite->start_tag_hash;
                hitem = tag_hash->lookup(tag_hash, X->tagname(X));
                if(hitem){
                    hitem->handler(X);
                    break;
                }

                /* default start_element handler: */
                if(X->xpin->suite->start_element_handler)
                    X->xpin->suite->start_element_handler(X);
                break;
            }
        case XPIN_END_ELEMENT:
            {
                /* if namespace,  ns_end_element handler: */
                struct xpin_node *N = X->node(X);

                if((X->xpin->ns_parser != 0)
                    && (N->ns_uri(N) !=  NULL)  
                    && (X->xpin->suite->ns_end_element_handler))
                {
                    X->xpin->suite->ns_end_element_handler(X);
                    break;
                }

                /* test for tagpath in end_tagpath_list: */
                tagpath_list = X->xpin->suite->tagpath_list;
                hitem = tagpath_list->search(
                          tagpath_list,
                          &xpin_tagpath_match,
                          X->tagpath(X));
                if(hitem){
                    hitem->handler(X);
                    break;
                }

                /* test for tag in end_tag_hash: */
                tag_hash = X->xpin->suite->end_tag_hash;
                hitem = tag_hash->lookup(tag_hash, X->tagname(X));
                if(hitem){
                    hitem->handler(X);
                    break;
                }

                if(X->xpin->suite->end_element_handler)
                    X->xpin->suite->end_element_handler(X);
                break;
            }
        case XPIN_PI:
            {
                dt_Hash          pi_hash = X->xpin->suite->pi_hash;
                struct xpin_pi  *pi = X->xpin->pi;

                hitem = pi_hash->lookup(pi_hash, pi->target(pi));
                if(hitem){
                    hitem->handler(X);
                    break;
                }

                if(X->xpin->suite->pi_handler)
                    X->xpin->suite->pi_handler(X);
                break;
            }
        case XPIN_START_CDATA:
            {
                if(X->xpin->suite->start_cdata_handler)
                    X->xpin->suite->start_cdata_handler(X);
                break;
            }
        case XPIN_END_CDATA:
            {
                if(X->xpin->suite->end_cdata_handler)
                    X->xpin->suite->end_cdata_handler(X);
                break;
            }
        default:
            {
                /* shouldn't be here! */
                assert(1);
                return -1;
            }
    }/*end switch*/

    return 0;
}/*end xpin_handler_dispatch()*/



/* that's all, folks! */
