/* xpin_node.c
**
** XpinDoc
** an XML parse instance document processing utility
**
** Copyright (C) 2002, 2003, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.09.29
** ===
*/
#include <ctype.h>
#include <fnmatch.h>
#include <stdlib.h>
#include <string.h>

/* our headers: */
#include "xpindoc.h"
#include "xpin_text.h"
#include "xpin_node.h"

/* node object internals: */
struct xpin_node_ {
    int                 depth;
    const char         *tagname;
    const char         *tagpath;
    const char         *local_name;
    const char         *ns_uri;
    const char         *ns_prefix;
    const char         *qname;
    int                 attr_count;
    struct xpin_attr  **attributes;
    struct xpin_node   *parent;
    struct xpin_text   *content;
};


/* internal prototypes: */

/* NODE METHODS: */
static
int
node_get_depth(struct xpin_node *);

static
const char *
node_get_tagname(struct xpin_node *);

static
const char *
node_get_tagpath(struct xpin_node *);

static
const char *
node_get_local_name(struct xpin_node *);

static
const char *
node_get_ns_uri(struct xpin_node *);

static
const char *
node_get_ns_prefix(struct xpin_node *);

static
const char *
node_get_qname(struct xpin_node *);

static
int
node_tagpath_match(struct xpin_node *, const char *pattern);

static
int
node_get_attr_count(struct xpin_node *);

static
const char *
node_get_attr_value(struct xpin_node *, const char *);

static
struct xpin_attr **
node_get_attributes(struct xpin_node *);

static
struct xpin_node *
node_get_parent(struct xpin_node *);

static
const char *
node_get_content(struct xpin_node *);

static
struct xpin_text *
node_get_text(struct xpin_node *);

/* NODE CONTENT METHODS: */
static
size_t
node_content_length(struct xpin_node *);

static
const char *
node_content_lower(struct xpin_node *);

static
const char *
node_content_upper(struct xpin_node *);

static
const char *
node_content_ltrim(struct xpin_node *);

static
const char *
node_content_rtrim(struct xpin_node *);

static
const char *
node_content_trim(struct xpin_node *);

static
const char *
node_content_chomp(struct xpin_node *);

static
const char *
node_content_normalize(struct xpin_node *);

static
const char *
node_content_append(struct xpin_node *, const char *);

static
const char *
node_content_prepend(struct xpin_node *, const char *);

static
const char *
node_content_merge(struct xpin_node *);



/* node method definitions: */
static
int
node_get_depth(struct xpin_node *N)
{
    return N->node->depth;
}


static
const char *
node_get_tagname(struct xpin_node *N)
{
    return N->node->tagname;
}


static
const char *
node_get_tagpath(struct xpin_node *N)
{
    return N->node->tagpath;
}


static
const char *
node_get_local_name(struct xpin_node *N)
{
    return N->node->local_name;
}


static
const char *
node_get_ns_uri(struct xpin_node *N)
{
    return N->node->ns_uri;
}


static
const char *
node_get_ns_prefix(struct xpin_node *N)
{
    return N->node->ns_prefix;
}


static
const char *
node_get_qname(struct xpin_node *N)
{
    return N->node->qname;
}


static
int
node_tagpath_match(struct xpin_node *N, const char *pattern)
{
    return fnmatch(N->node->tagpath, pattern, 0);
}


static
int
node_get_attr_count(struct xpin_node *N)
{
    return N->node->attr_count;
}


static
const char *
node_get_attr_value(struct xpin_node *N, const char *key)
{
    struct xpin_attr   **attributes = NULL;
    int                  i = 0;

    if(!N || !N->node)  return NULL;

    attributes = N->node->attributes;
    if(!attributes || !key)  return NULL;

    /* sequential lookup for key in attribute->name: */
    for(i = 0; i < N->node->attr_count; ++i){
        if((strcmp(key, attributes[i]->name)) == 0){
            return attributes[i]->value;
        }
    }

    /* not found: */
    return NULL;
}


static
struct xpin_attr **
node_get_attributes(struct xpin_node *N)
{
    return N->node->attributes;
}



static
struct xpin_node *
node_get_parent(struct xpin_node *N)
{
    return N->node->parent;
}


static
const char *
node_get_content(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;

    return T->content(T);
}

static
struct xpin_text *
node_get_text(struct xpin_node *N)
{
    return N->node->content;
}


static
size_t
node_content_length(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;

    return T->length(T);
}


static
const char *
node_content_lower(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;

    return T->to_lower(T);
}

    
static
const char *
node_content_upper(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;

    return T->to_upper(T);
}
    


static
const char *
node_content_ltrim(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;

    return T->ltrim(T);
}


static
const char *
node_content_rtrim(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;

    return T->rtrim(T);
}


static
const char *
node_content_trim(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;

    return T->trim(T);
}


static
const char *
node_content_chomp(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;

    return T->chomp(T);
}


static
const char *
node_content_normalize(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;

    return T->normalize(T);
}/*end node_content_normalize()*/



static
const char *
node_content_append(struct xpin_node *N, const char *from)
{
    struct xpin_text  *T = N->node->content;

    return T->append(T, from);
}/*end node_content_append()*/


static
const char *
node_content_prepend(struct xpin_node *N, const char *from)
{
    struct xpin_text    *T = N->node->content;

    return T->prepend(T, from);
}/*end node_content_prepend()*/


static
const char *
node_content_merge(struct xpin_node *N)
{
    struct xpin_text  *T = N->node->content;
    struct xpin_node  *parent = N->parent(N);

    return node_content_append(parent, T->content(T));
}/*end node_content_merge()*/



/* exported (xpin_node.h): */

/* NODE CONSTRUCTOR: */

struct xpin_node *
new_xpin_node
(
    int                 depth,
    const char         *tagname,
    const char         *tagpath,
    const char         *local_name,
    const char         *ns_uri,
    const char         *ns_prefix,
    const char         *qname,
    int                 attr_count,
    struct xpin_attr  **attributes,
    struct xpin_node   *parent
)
{
    struct xpin_node   *N = NULL;
    struct xpin_node_  *node = NULL;
    struct xpin_text   *T = NULL;

    N = (struct xpin_node *)malloc(sizeof(struct xpin_node));
    if(N == NULL)  return NULL;

    node = (struct xpin_node_ *)malloc(sizeof(struct xpin_node_));
    if(node == NULL){
        free(N);
        return NULL;
    }

    T = new_xpin_text();
    if(T == NULL){
        free(node);
        free(N);
        return NULL;
    }

    node->depth      = depth;
    node->tagname    = tagname;
    node->tagpath    = tagpath;
    node->local_name = local_name;
    node->ns_uri     = ns_uri;
    node->ns_prefix  = ns_prefix;
    node->qname      = qname;
    node->attr_count = attr_count;
    node->attributes = attributes;
    node->parent     = parent;
    node->content    = T;

    N->node = node;

    /* methods: */
    N->depth         = &node_get_depth;
    N->tagname       = &node_get_tagname;
    N->tagpath       = &node_get_tagpath;
    N->local_name    = &node_get_local_name;
    N->ns_uri        = &node_get_ns_uri;
    N->ns_prefix     = &node_get_ns_prefix;
    N->qname         = &node_get_qname;
    N->tagpath_match = &node_tagpath_match;
    N->attr_count    = &node_get_attr_count;
    N->attr_value    = &node_get_attr_value;
    N->attributes    = &node_get_attributes;
    N->parent        = &node_get_parent;
    N->content       = &node_get_content;
    N->text          = &node_get_text;

    N->content_length     = &node_content_length;
    N->content_lower      = &node_content_lower;
    N->content_upper      = &node_content_upper;
    N->content_ltrim      = &node_content_ltrim;
    N->content_rtrim      = &node_content_rtrim;
    N->content_trim       = &node_content_trim;
    N->content_chomp      = &node_content_chomp;
    N->content_normalize  = &node_content_normalize;
    N->content_append     = &node_content_append;
    N->content_prepend    = &node_content_prepend;
    N->content_merge      = &node_content_merge;
    /* ... */

    return N;
}/*end new_xpin_node()*/



/* NODE DESTRUCTOR: */
void
free_xpin_node(struct xpin_node *N)
{
    struct xpin_text  *T = NULL;
    int   i = 0;

    if(!N || !N->node)  return;

    /* DON'T free tagname, it is either local_name or qname! */
    if(N->node->tagpath)    free((char *)N->node->tagpath);
    if(N->node->local_name) free((char *)N->node->local_name);
    if(N->node->ns_uri)     free((char *)N->node->ns_uri);
    if(N->node->ns_prefix)  free((char *)N->node->ns_prefix);
    if(N->node->qname)      free((char *)N->node->qname);

    /* free attributes: */
    for(i = 0; i < N->node->attr_count; ++i){
        /* attributes name is qname: */
        if(N->node->attributes[i]->name)
            free((char *)N->node->attributes[i]->name);
        if(N->node->attributes[i]->local_name)
            free((char *)N->node->attributes[i]->local_name);
        if(N->node->attributes[i]->ns_uri)
            free((char *)N->node->attributes[i]->ns_uri);
        if(N->node->attributes[i]->ns_prefix)
            free((char *)N->node->attributes[i]->ns_prefix);
        if(N->node->attributes[i]->value)
            free((char *)N->node->attributes[i]->value);
        free(N->node->attributes[i]);
    }
    if(N->node->attributes)  free(N->node->attributes);

    /*
    ** N->node->parent not freed, as it will continue
    ** to be referenced when current node is popped
    */

    /* free content (xpin_text object): */
    if((T = N->node->content) != NULL)
        T->free(T);

    /* free the rest: */
    free(N->node);
    free(N);

    return;
}/*end free_xpin_node()*/



/* that's all, folks! */
