/* xpin_text.c
**
** XpinDoc
** an XML parse instance document processing utility
**
** Copyright (C) 2002, 2003, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.09.30
** ===
*/
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* our headers: */
#include "xpindoc.h"
#include "xpin_text.h"

#define  TEXT_INIT_SIZE  1


/*
** "data" is stored as a nul-terminated string
** "cur_used" includes the terminating nul
**   -- ie, cur_used = strlen(data) + 1;
** "data" is guaranteed to be at least the nul string "\0"
**   -- ie, cur_used >= 1
*/


/* growable array for content: */
struct xpin_text_ {
    int     cur_used;
    int     cur_alloc;
    char   *data;
};



/* internal prototypes: */

/* xpin_text dynamic grow: */
static
char *
xpin_text_grow(struct xpin_text *, int);


/* xpin_text methods: */
static
const char *
xpin_text_get_data(struct xpin_text *);

static
size_t
xpin_text_get_length(struct xpin_text *);

static
int
xpin_text_clear(struct xpin_text *);

static
const char *
xpin_text_to_lower(struct xpin_text *);

static
const char *
xpin_text_to_upper(struct xpin_text *);

static
const char *
xpin_text_ltrim(struct xpin_text *);

static
const char *
xpin_text_rtrim(struct xpin_text *);

static
const char *
xpin_text_trim(struct xpin_text *);

static
const char *
xpin_text_chomp(struct xpin_text *);

static
const char *
xpin_text_normalize(struct xpin_text *);

static
const char *
xpin_text_prepend(struct xpin_text *, const char *);

static
const char *
xpin_text_put_char(struct xpin_text *, char);

static
const char *
xpin_text_put_string(struct xpin_text *T, const char *string);

static
const char *
xpin_text_put_chardata(struct xpin_text *T, const char *data, int len);

static
int
xpin_text_printf(struct xpin_text *T, const char *format, ...);

/* destructor method: */
static
void
xpin_text_free(struct xpin_text *T);


/* DEFINITIONS: */

/* xpin_text creator: */
struct xpin_text *
new_xpin_text(void)
{
    struct xpin_text    *T = NULL;
    struct xpin_text_   *text = NULL;
    char                *data = NULL;

    T = (struct xpin_text *)malloc(sizeof(struct xpin_text));
    if(!T)  return NULL;

    text = (struct xpin_text_ *)malloc(sizeof(struct xpin_text_));
    if(!text){
        free(T);
        return NULL;
    }

    data = (char *)malloc(TEXT_INIT_SIZE * sizeof(char));
    if(!data){
        free(text);
        free(T);
        return NULL;
    }

    /* guarantee nul terminated data: */
    data[0] = '\0';

    text->data = data;
    text->cur_used = 1;
    text->cur_alloc = TEXT_INIT_SIZE;

    T->text = text;

    /* methods: */
    T->content      = &xpin_text_get_data;
    T->length       = &xpin_text_get_length;
    T->clear        = &xpin_text_clear;

    T->to_lower     = &xpin_text_to_lower;
    T->to_upper     = &xpin_text_to_upper;

    T->ltrim        = &xpin_text_ltrim;
    T->rtrim        = &xpin_text_rtrim;
    T->trim         = &xpin_text_trim;
    T->chomp        = &xpin_text_chomp;
    T->normalize    = &xpin_text_normalize;
    T->append       = &xpin_text_put_string;
    T->prepend      = &xpin_text_prepend;

    T->put_char     = &xpin_text_put_char;
    T->put_string   = &xpin_text_put_string;
    T->put_chardata = &xpin_text_put_chardata;
    T->printf       = &xpin_text_printf;

    T->free         = &xpin_text_free;

    return T;   
}/*end new_xpin_text()*/


/* xpin_text destructor: */
static
void
xpin_text_free(struct xpin_text *T)
{
    if(T){
        if(T->text){
           if(T->text->data) free(T->text->data);
           free(T->text);
        }
        free(T);
    }
    return;
}/*end free_xpin_text()*/



static
char *
xpin_text_grow(struct xpin_text *T, int add)
{
    struct xpin_text_   *text = T->text;
    int                  n = text->cur_used + add;

    /* grow data if cur_alloc < cur_used + add */
    if(text->cur_alloc < n){
        int     nn = text->cur_alloc * 2;
        char   *cp = NULL;

        /* grow array at least double: */
        if(n < nn)
            n = nn;

        cp = (char *)realloc(text->data, n);
        if(!cp)
            /* XXX error! */
            return NULL;
        else
            text->data = cp;
            text->cur_alloc = n;
    }

    return text->data;
}/*end xpin_text_grow()*/



static
const char *
xpin_text_get_data(struct xpin_text *T)
{
    return (const char *)T->text->data;
}/*end xpin_text_get_data()*/



static
size_t
xpin_text_get_length(struct xpin_text *T)
{
    return strlen(T->text->data);
}/*end xpin_text_get_length()*/



static
int
xpin_text_clear(struct xpin_text *T)
{
    T->text->data[0] = '\0';
    T->text->cur_used = 1;

    return 0;
}



static
const char *
xpin_text_to_lower(struct xpin_text *T)
{
    char  *t = T->text->data;

    while(*t){
        *t = tolower((int)*t);
        ++t;
    }

    return (const char *)T->text->data;
}/*end xpin_text_to_lower()*/




static
const char *
xpin_text_to_upper(struct xpin_text *T)
{
    char  *t = T->text->data;

    while(*t){
        *t = toupper((int)*t);
        ++t;
    }

    return (const char *)T->text->data;
}/*end xpin_text_to_upper()*/



static
const char *
xpin_text_ltrim(struct xpin_text *T)
{
    char              *t0 = T->text->data;
    char              *tN = t0;

    while(*tN && isspace((int)*tN)){
        ++tN;
        --T->text->cur_used;
    }

    if(tN != t0){
        if(*tN != '\0'){
            while(*tN)
                *t0++ = *tN++;
        }
        *t0 = '\0';
    } 

    return (const char *)T->text->data;
}/*end xpin_text_ltrim()*/



static
const char *
xpin_text_rtrim(struct xpin_text *T)
{
    char   *t0 = T->text->data;
    char   *tN = NULL;
    size_t  n = 0;

    n = strlen(t0);
    if(n < 1) return (const char *)t0;

    tN = t0 + n - 1;
    while( (tN >= t0) && (isspace((int)*tN)) ){
        *tN-- = '\0';
        --T->text->cur_used;
    }

    return (const char *)t0;
}/*end xpin_text_rtrim()*/



static
const char *
xpin_text_trim(struct xpin_text *T)
{
    xpin_text_rtrim(T);
    return xpin_text_ltrim(T);
}/*end xpin_text_trim()*/



static
const char *
xpin_text_chomp(struct xpin_text *T)
{
    char   *t0 = T->text->data;
    char   *tN = NULL;
    size_t  n = 0;

    n = strlen(t0);
    if(n < 1) return (const char *)T->text->data;

    tN = t0 + n - 1;
    while( (tN >= t0)
        && ((*tN == '\n') || (*tN == '\r')) )
    {
        *tN-- = '\0';
        --T->text->cur_used;
    }

    return (const char *)t0;
}/*end xpin_text_chomp()*/



static
const char *
xpin_text_normalize(struct xpin_text *T)
{
    char              *t0 = NULL, *t1 = NULL;

    /* first trim leading and trailing: */
    xpin_text_trim(T);

    t0 = t1 = T->text->data;

    /* reduce any sequence of two or more "spaces" to single space: */
    while(*t1){
        if(isspace((int)*t1)){
            /* convert any "space" char to a space: */
            *t0++ = 040;
            /* increment through additional spaces: */
            ++t1;
            while(*t1 && (isspace((int)*t1))){
                ++t1;
                --T->text->cur_used;
            }
        }
        else
            *t0++ = *t1++;
    }
    *t0 = '\0';

    return (const char *)T->text->data;
}/*end xpin_text_normalize()*/



static
const char *
xpin_text_prepend(struct xpin_text *T, const char *string)
{
    char                *t0 = NULL, *t1 = NULL, *tN = NULL;
    size_t               len = 0;

    if(!string) return (const char *)T->text->data;

    len = strlen(string);

    t0 = xpin_text_grow(T, len);
    if(!t0)
        /* XXX error! */
        return NULL;

    t1 = t0 + T->text->cur_used - 1;
    tN = t1 + len;

    /* move current text "forward" */
    while(t1 >= t0)
        *tN-- = *t1--; 
   
    /* copy in new text: */
    while(*string)
        *t0++ = *string++;

    T->text->cur_used += len;

    return (const char *)T->text->data; 
}/*end xpin_text_prepend()*/


static
const char *
xpin_text_put_char(struct xpin_text *T, char c)
{
    struct xpin_text_   *text = T->text;
    char                *t = NULL;

    t = xpin_text_grow(T, 1);
    if(!t)
        /* XXX error! */
        return NULL;

    t += text->cur_used - 1;
    *t++ = c;
    *t = '\0';
    ++text->cur_used; 

    return (const char *)text->data;        
}/*end xpin_text_put_char()*/



static
const char *
xpin_text_put_string(struct xpin_text *T, const char *string)
{
    struct xpin_text_   *text = T->text;
    char                *t = NULL;
    int                  n = 0;
    char                *s = (char *)string;

    if(!string)
        /*XXX error? */
        return (const char *)text->data;

    n = strlen(string);
    t = xpin_text_grow(T, n);
    if(!t)
        /* XXX error! */
        return NULL;

    t += text->cur_used - 1;

    while(*s){
        *t++ = *s++;
    }
    *t = '\0';
    text->cur_used += n;

    return (const char *)text->data;
}/*end xpin_text_put_string()*/



static
const char *
xpin_text_put_chardata
(
    struct xpin_text  *T,
    const char        *chardata,
    int                len
)
{
    struct xpin_text_   *text = T->text;
    char                *t = NULL;
    const char          *s = chardata;
    int                  i = 0;

    if(!chardata || (len == 0))
        return (const char *)text->data;

    t = xpin_text_grow(T, len);
    if(!t)
        /* XXX error! */
        return NULL;

    t += text->cur_used - 1;
    for(i = 0; i < len; ++i){
        *t++ = *s++;
    }
    *t = '\0';
    text->cur_used += len;

    return (const char *)text->data;
}/*end xpin_text_put_chardata()*/



static
int
xpin_text_printf(struct xpin_text *T, const char *format, ...)
{
    va_list              ap;
    struct xpin_text_   *text = T->text;
    char                *s = NULL, *s0 = NULL;
    char                *t = NULL;
    int                  n = 0;

    if(!format) return 0;

    va_start(ap, format);
    n = vasprintf(&s, format, ap);
    va_end(ap);

    if((n < 0) || (s == NULL)){
        /* XXX error in vasprintf()! */
        if(s)  free(s);
        return -1;
    }

    n = strlen(s);
    t = xpin_text_grow(T, n);
    if(!t){
        /* XXX error! */
        free(s);
        return -1;
    }

    t += text->cur_used - 1;

    s0 = s;
    while(*s0){
        *t++ = *s0++;
    }
    *t = '\0';

    text->cur_used += n;
    free(s);

    return 0;
}/*end xpin_text_printf()*/



/* that's all, folks! */
