/* xpin_writer.c
**
** XpinDoc
** an XML parse instance document processing utility
**
** Copyright (C) 2002, 2003, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.09.19
** ===
*/
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* our headers: */
#include "xpindoc.h"
#include "xpin.h"
#include "xpin_writer.h"
#include "dt_list.h"

/* opaque internals: */
struct xpin_writer_ {
    /* access xpindoc internals: */
    struct xpin  *xpin;
    /* output stack: */
    dt_List       output_stack;
};

struct output_record {
    void         *dest;
    int           dest_type;
};

/* internal prototypes: */
static
struct output_record *
new_output_record(void *dest, int dest_type);

static
void
free_output_record(struct output_record *);

/* internal xpinout sub-routines (using output_stack): */
static
int
xpinout_put_char(struct xpin_writer *W, char c);

static
int
xpinout_put_string(struct xpin_writer *W, const char *string);

static
int
xpinout_put_chardata(struct xpin_writer *W, const char *data, int len);

/* xpin_writer methods: */
static
int
xpin_writer_put_char(struct xpin_writer *W, const char);

static
int
xpin_writer_put_string(struct xpin_writer *W, const char *string);

static
int
xpin_writer_put_chardata(struct xpin_writer *W, const char *data, int len);

static
int
xpin_writer_printf(struct xpin_writer *W, const char *format, ...);

static
int
xpin_writer_push_output(struct xpin_writer *W, void *dest, int dest_type);

static
void *
xpin_writer_pop_output(struct xpin_writer *W);



/* DEFINITIONS: */

/* xpin_writer creator: */
struct xpin_writer *
new_xpin_writer(struct xpin *xpin)
{
    struct xpin_writer    *W = NULL;
    struct xpin_writer_   *writer = NULL;
    dt_List                output_stack = NULL;

    W = (struct xpin_writer *)malloc(sizeof(struct xpin_writer));
    if(!W)  return NULL;

    writer = (struct xpin_writer_ *)malloc(sizeof(struct xpin_writer_));
    if(!writer){
        free(W);
        return NULL;
    }

    output_stack = new_list();
    if(!output_stack){
        free(writer);
        free(W);
        return NULL;
    }

    /* note:
       initialization of default output_stack (record #1) is
       handled in initialization of xpindoc object
    */

    writer->output_stack = output_stack;

    writer->xpin = xpin;

    W->writer = writer;

    /* methods: */
    W->put_char     = &xpin_writer_put_char;
    W->put_string   = &xpin_writer_put_string;
    W->put_chardata = &xpin_writer_put_chardata;
    W->printf       = &xpin_writer_printf;
    W->push_output  = &xpin_writer_push_output;
    W->pop_output   = &xpin_writer_pop_output;
    return W;   
}/*end new_xpin_writer()*/


/* xpin_writer destructor: */
void
free_xpin_writer(struct xpin_writer *W)
{
    if(W){
        if(W->writer){
           /* XXX free writer->output_stack! */
           free(W->writer);
        }
        free(W);
    }
    return;
}/*end free_xpin_writer()*/



static
struct output_record *
new_output_record(void *dest, int dest_type)
{
    struct output_record  *rec = NULL;

    rec = (struct output_record *)malloc(sizeof(struct output_record));
    if(!rec)  return NULL;

    rec->dest = dest;
    rec->dest_type = dest_type;
    return rec;
}/*end new_output_record()*/



static
void
free_output_record(struct output_record *rec)
{
    if(rec != NULL) free(rec);
    return;
}/*end free_output_record()*/ 



static
int
xpinout_put_char(struct xpin_writer *W, char c)
{
    dt_List                 output_stack = W->writer->output_stack;
    struct output_record   *rec = NULL;
    int                     dest_type = 0;

    rec = (struct output_record *)output_stack->peek(output_stack);
    dest_type = rec->dest_type;
    switch(dest_type){
        case XPINOUT_PIPE:  /* pass through */
        case XPINOUT_FILE:
        {
            FILE *F = (FILE *)rec->dest;
            int   ferr = fputc(c, F);
            if(ferr == EOF){
                /* XXX error! */
            }
            break;
        }
        case XPINOUT_TEXT:
        {
            struct xpin_text *T =
                (struct xpin_text *)rec->dest;

            T->put_char(T, c);
            break;
        }
        case XPINOUT_NULL:
        {
            /* do nothing */
            break;
        }
        default:
        {
            /* XXX not here! */
        }
    }

    /* XXX return value! */
    return 0;
}/*end xpinout_put_char()*/


static
int
xpinout_put_string(struct xpin_writer *W, const char *string)
{
    dt_List                 output_stack = W->writer->output_stack;
    struct output_record   *rec = NULL;
    int                     dest_type = 0;

    rec = (struct output_record *)output_stack->peek(output_stack);
    dest_type = rec->dest_type;
    switch(dest_type){
        case XPINOUT_PIPE:  /* pass through */
        case XPINOUT_FILE:
        {
            FILE *F = (FILE *)rec->dest;
            int   ferr = fprintf(F, "%s", string);
            if(ferr < 0){
                /* XXX error! */
            }
            break;
        }
        case XPINOUT_TEXT:
        {
            struct xpin_text *T =
                (struct xpin_text *)rec->dest;

            T->put_string(T, string);
            break;
        }
        case XPINOUT_NULL:
        {
            /* do nothing */
            break;
        }
        default:
        {
            /* XXX not here! */
        }
    }

    /* XXX return value! */
    return 0;
}/*end xpinout_put_string()*/



static
int
xpinout_put_chardata(struct xpin_writer *W, const char *data, int len)
{
    dt_List                 output_stack = W->writer->output_stack;
    struct output_record   *rec = NULL;
    int                     dest_type = 0;

    rec = (struct output_record *)output_stack->peek(output_stack);
    dest_type = rec->dest_type;
    switch(dest_type){
        case XPINOUT_PIPE:  /* pass through */
        case XPINOUT_FILE:
        {
            FILE *F = (FILE *)rec->dest;
            int   ferr;
            while(len){
                ferr = fputc(*data, F);
                if(ferr == EOF){
                    /* XXX error! */
                    break;
                }
                ++data;
                --len;
            }
            break;
        }
        case XPINOUT_TEXT:
        {
            struct xpin_text *T =
                (struct xpin_text *)rec->dest;

            T->put_chardata(T, data, len);
            break;
        }
        case XPINOUT_NULL:
        {
            /* do nothing */
            break;
        }
        default:
        {
            /* XXX not here! */
        }
    }

    /* XXX return value! */
    return 0;
}/*end xpinout_put_chardata()*/




static
int
xpin_writer_put_char(struct xpin_writer *W, char c)
{
    struct xpin_writer_  *writer = W->writer;
    struct xpin          *xpin = writer->xpin;
    int                   datamode = xpin->datamode;

    if(datamode == XPIN_CAPTURE){
        struct xpin_node  *node = xpin->node;
        struct xpin_text  *T = node->text(node);
        T->put_char(T, c);
    }
    else if(datamode == XPIN_OUTPUT){
        xpinout_put_char(W, c);
    }
    else{
        /* XXX error, unknown datamode */
    }

    /* XXX return value? */
    return 0;
}/*end xpin_writer_put_char()*/



static
int
xpin_writer_put_string(struct xpin_writer *W, const char *string)
{
    struct xpin_writer_  *writer = W->writer;
    struct xpin          *xpin = writer->xpin;
    int                   datamode = xpin->datamode;

    if(datamode == XPIN_CAPTURE){
        struct xpin_node  *node = xpin->node;
        struct xpin_text  *T = node->text(node);
        T->put_string(T, string);
    }
    else if(datamode == XPIN_OUTPUT){
        xpinout_put_string(W, string);
    }
    else{
        /* XXX error, unknown datamode */
        fprintf(stderr, "xpin_writer_put_string(): unknown datamode\n");
    }

    /* XXX return value? */
    return 0;
}/*end xpin_writer_put_string()*/



static
int
xpin_writer_put_chardata
(
    struct xpin_writer  *W,
    const char          *chardata,
    int                  len
)
{
    struct xpin_writer_  *writer = W->writer;
    struct xpin          *xpin = writer->xpin;
    int                   datamode = xpin->datamode;

    if(datamode == XPIN_CAPTURE){
        struct xpin_node  *node = xpin->node;
        struct xpin_text  *T = node->text(node);
        T->put_chardata(T, chardata, len);
    }
    else if(datamode == XPIN_OUTPUT){
        xpinout_put_chardata(W, chardata, len);
    }
    else{
        /* XXX error, unknown datamode */
    }

    /* XXX return value? */
    return 0;
}/*end xpin_writer_put_chardata()*/



static
int
xpin_writer_printf(struct xpin_writer *W, const char *format, ...)
{
    va_list               ap;
    struct xpin_writer_  *writer = W->writer;
    struct xpin          *xpin = writer->xpin;
    int                   datamode = xpin->datamode;
    char                 *s = NULL;
    int                   err = 0;

    va_start(ap, format);
    err = vasprintf(&s, format, ap);
    va_end(ap);
    if((err < 0) || (s == NULL)){
        /* XXX error with vasprintf()! */
        if(s)  free(s);
        return err;
    }

    if(datamode == XPIN_CAPTURE){
        struct xpin_node  *node = xpin->node;
        struct xpin_text  *T = node->text(node);
        T->put_string(T, s);
    }
    else if(datamode == XPIN_OUTPUT){
        xpinout_put_string(W, s);
    }
    else{
        /* XXX error, unknown datamode */
    }

    if(s)  free(s);

    /* XXX return value? */
    return 0;
}/*end xpin_writer_printf()*/



static
int
xpin_writer_push_output(struct xpin_writer *W, void *dest, int dest_type)
{
    struct output_record  *rec = NULL;
    dt_List                output_stack = W->writer->output_stack;

    rec = new_output_record(dest, dest_type);
    if(!rec){
        /* XXX error! */
        fprintf(stderr, "error: null rec in xpinout_push()\n");
        return -1;
    }
    output_stack->push(output_stack, (void *)rec);
    return 0;
}/*end xpinout_push()*/



static
void *
xpin_writer_pop_output(struct xpin_writer *W)
{
    dt_List                 output_stack = W->writer->output_stack;
    struct output_record   *rec = NULL;
    void                   *dest = NULL;

    /* don't pop last item! */
    if(output_stack->items(output_stack) > 1){
        rec = output_stack->pop(output_stack);
        if(rec->dest_type == XPINOUT_TEXT){
            dest = rec->dest;
        }
        else{
            dest = NULL;
        }
        free_output_record(rec);
    }
    return (void *)dest;
}/*end xpinout_pop()*/




/* that's all, folks! */
