/* xpincap.c
** XpinDoc "xpincap"
**
** stress test for XpinDoc:
** capture all content and merge up to root
**
** "cat" an xml file with xpindoc in capture mode
**
** wcm, 2003.09.23 - 2003.10.01
** ===
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "xpindoc.h"
#include "xpin_macros.h"

#define  SBUF_SIZE  512

/* xpincap prototypes: */
void
my_set_handler(XpinDoc X, const char *pattern, Xpin_Handler callback);

void
xml_filter(XpinDoc X, const char *data, int len);

void
xpincap_start_document(XpinDoc xpin);

void
xpincap_end_document(XpinDoc xpin);

void
xpincap_start_element(XpinDoc xpin);

void
xpincap_end_element(XpinDoc xpin);

void
xpincap_pi(XpinDoc xpin);

void
xpincap_start_cdata(XpinDoc xpin);

void
xpincap_end_cdata(XpinDoc xpin);


/* definitions: */

/* error trapping wrapper for set_handler(): */
void
my_set_handler(XpinDoc X, const char *pattern, Xpin_Handler callback)
{
    int   err = 0;

    err = X->set_handler(X, pattern, callback);
    if(err){
        fprintf(stderr, "xpinman error in: my_set_handler(X, %s, callback)\n",
                pattern);
        exit(EXIT_FAILURE);
    }

    return;
}/*end my_set_handler()*/


void
xml_filter(XpinDoc X, const char *data, int len)
{
    char       *c = (char *)data;

    while(len){
        if(*c == '<') X->put_string(X, "&lt;");
        else if(*c == '>') X->put_string(X, "&gt;");
        else if(*c == '&') X->put_string(X, "&amp;");
        else X->put_char(X, *c);
        ++c;
        --len;
    }

    return;
}


void
xpincap_start_document(XpinDoc xpin)
{
    /* do nothing: */
    return;
}


void
xpincap_end_document(XpinDoc xpin)
{
    /* output all content merged into root element: */
    printf("%s\n", xpin->content(xpin));

    return;
}


void
xpincap_start_element(XpinDoc xpin)
{
    Xpin_Node        N = xpin->node(xpin);
    Xpin_Attributes  A = N->attributes(N);
    int              attr_count = N->attr_count(N);
    int              i = 0;
    char             sbuf[SBUF_SIZE];

    xpin->put_char(xpin, '<');
    xpin->put_string(xpin, xpin->tagname(xpin));

    for(i = 0; i < attr_count; ++i){
        /* output attributes: */
        snprintf(sbuf, sizeof(sbuf),
                 " %s=\"%s\"",
                 A[i]->name, A[i]->value );

        xpin->put_string(xpin, sbuf);
    }
 
    xpin->put_char(xpin, '>');

    return;
}

void
xpincap_end_element(XpinDoc xpin)
{
    Xpin_Node  N = xpin->node(xpin);

    xpin->put_string(xpin, "</");
    xpin->put_string(xpin, xpin->tagname(xpin));
    xpin->put_char(xpin, '>');

    /* merge content into parent: */
    N->content_merge(N);

    return;
}


void
xpincap_pi(XpinDoc xpin)
{
    Xpin_PI   pi = xpin->pi(xpin);

    xpin->printf(xpin, "<?%s %s?>", pi->target(pi), pi->data(pi));
    return;
}



void
xpincap_start_cdata(XpinDoc xpin)
{
    xpin->put_string(xpin, "<![CDATA[");
    xpin->set_filter(xpin, NULL);
    return;
}



void
xpincap_end_cdata(XpinDoc xpin)
{
    xpin->put_string(xpin, "]]>");
    xpin->set_filter(xpin, &xml_filter);
    return;
}




int
xpindoc_main(int argc, char **argv)
{
    FILE    *input = NULL;
    XpinDoc  xpin = NULL;
    int      err = 0;
    char    *mydata = NULL;

    input = stdin;

    xpin = new_xpindoc(XPIN_NAMESPACE | XPIN_QNAME);
    if(xpin == NULL){
        fprintf(stderr, "error creating xpindoc\n");
        exit(EXIT_FAILURE);
    }

    mydata = strdup("This is xpincap!");

    xpin->set_mydata(xpin, (void *)mydata);

    xpin->set_datamode(xpin, XPIN_CAPTURE);
    xpin->set_filter(xpin, &xml_filter);

    my_set_handler(xpin, "Start_Document", &xpincap_start_document);
    my_set_handler(xpin, "End_Document", &xpincap_end_document);
    my_set_handler(xpin, "Start_Element", &xpincap_start_element);
    my_set_handler(xpin, "End_Element", &xpincap_end_element);
    my_set_handler(xpin, "Processing_Instruction", &xpincap_pi);
    my_set_handler(xpin, "Start_CDATA", &xpincap_start_cdata);
    my_set_handler(xpin, "End_CDATA", &xpincap_end_cdata);

    err = xpin->parse_stream(xpin, input);
    if(err){
        fprintf(stderr, "error during parse\n");
        exit(EXIT_FAILURE);
    }

    free(mydata);
    xpin->free(xpin);

    return 0;
}/*end xpindoc_main()*/


int
main(int argc, char **argv)
{
    return xpindoc_main(argc, argv);
}

/* that's all, folks! */
