/* xpindoc.h
**
** XpinDoc
** an XML parse instance document processing utility
**
** exported library api 
**
** Copyright (C) 2002, 2003, Wayne Marshall
** All rights reserved, worldwide.
**
** wcm, 2002.05.20 - 2003.09.30
** ===
*/
#ifndef  XPINDOC_H
#define  XPINDOC_H  1

#include <stdio.h>

enum xpin_parse_flags {
    XPIN_DEFAULT   = 0x0000,
    XPIN_NAMESPACE = 0x0001,
    XPIN_QNAME     = 0x0002
};


enum xpin_event_type {
    XPIN_NULL = 0,
    XPIN_START_DOCUMENT,
    XPIN_END_DOCUMENT,
    XPIN_START_ELEMENT,
    XPIN_END_ELEMENT,
    XPIN_PI,
    XPIN_START_CDATA,
    XPIN_END_CDATA,
    XPIN_COMMENT,
    XPIN_ERROR
};


enum xpin_datamode {
    XPIN_CAPTURE = 0,
    XPIN_OUTPUT  = 1
};
#define XPIN_DEFAULT_DATAMODE  XPIN_CAPTURE


enum xpin_ignore {
    XPIN_IGNORE_OFF = 0,
    XPIN_IGNORE_ON
};


enum xpin_parse_phase {
    XPIN_SETUP = 0,
    XPIN_PARSING,
    XPIN_DONE
};


enum xpinout_dest_type {
    XPINOUT_FILE = 0,
    XPINOUT_TEXT,
    XPINOUT_PIPE,
    XPINOUT_NULL
};


/* EVENTS: */
struct xpin_event {
    /* opaque internal representation: */
    struct xpin_event_     *event;
    /* public methods: */
    enum xpin_event_type  (*type)(struct xpin_event *);
    int                   (*line)(struct xpin_event *);
    int                   (*column)(struct xpin_event *);
    int                   (*size)(struct xpin_event *);
    long int              (*index)(struct xpin_event *);
    void                 *(*data)(struct xpin_event *);
    /* return data cast according to event type: */
    struct xpin_pi       *(*pi)(struct xpin_event *);
    char                 *(*comment)(struct xpin_event *);
};
typedef struct xpin_event*  Xpin_Event;


/* TEXT (for content in node object CAPTURE mode): */
/* also available for user content grabber with OUTPUT mode  */
struct xpin_text {
    /* opaque internal representation: */
    struct xpin_text_  *text;
    /* public methods: */
    const char         *(*content)(struct xpin_text *T);
    size_t              (*length)(struct xpin_text *T);
    int                 (*clear)(struct xpin_text *T);
    const char         *(*to_lower)(struct xpin_text *T);
    const char         *(*to_upper)(struct xpin_text *T);
    const char         *(*ltrim)(struct xpin_text *T);
    const char         *(*rtrim)(struct xpin_text *T);
    const char         *(*trim)(struct xpin_text *T);
    const char         *(*chomp)(struct xpin_text *T);
    const char         *(*normalize)(struct xpin_text *T);
    const char         *(*append)(struct xpin_text *T, const char *string);
    const char         *(*prepend)(struct xpin_text *T, const char *string);
    /* put_* methods _append_ to existing content: */
    const char         *(*put_char)(struct xpin_text *T, char c);
    const char         *(*put_string)(struct xpin_text *T, const char *string);
    const char         *(*put_chardata)
                            (
                                struct xpin_text  *T,
                                const char        *data,
                                int                len
                            );
    int                 (*printf)(struct xpin_text *T, const char *format, ...);
    /* destructor: */
    void                (*free)(struct xpin_text *T);
};
typedef struct xpin_text*  Xpin_Text;
/* export Xpin_Text object creator: */
extern Xpin_Text  new_xpin_text();


/* WRITER (for filtering input to output): */
/* includes stack-based output control for OUTPUT mode */
/* XXX only filter clients have direct access to writer */
struct xpin_writer {
    /* opaque internal representation: */
    struct xpin_writer_  *writer;
    /* public methods: */
    /*
       put_* methods differentiate between parser modes:

         mode           behavior
         -------------  ---------------------------------------------
         XPIN_CAPTURE:  write (append) to text object of current node
         XPIN_OUTPUT:   write to destination in head of output stack
    */
    int       (*put_char)(struct xpin_writer *W, char c);
    int       (*put_string)(struct xpin_writer *W, const char *string);
    int       (*put_chardata)
                  (
                      struct xpin_writer  *W,
                      const char          *data,
                      int                  len
                  );
    int       (*printf)(struct xpin_writer *W, const char *format, ...);
    /* stack-based output control methods: */
    int       (*push_output)(struct xpin_writer *W, void *dest, int dest_type);
    void     *(*pop_output)(struct xpin_writer *W); 

};
typedef struct xpin_writer* Xpin_Writer;



/* NODES: */
struct xpin_node {
    /* opaque internal representation: */
    struct xpin_node_  *node;
    /* public methods: */
    int                 (*depth)(struct xpin_node *);
    /* with XPIN_QNAME, tagname/tagpath uses qname, else local_name: */
    const char         *(*tagname)(struct xpin_node *);
    const char         *(*tagpath)(struct xpin_node *);
    int                 (*tagpath_match)(struct xpin_node *, const char *pattern);
    /* namespace support: */
    const char         *(*local_name)(struct xpin_node *);
    const char         *(*ns_uri)(struct xpin_node *);
    const char         *(*ns_prefix)(struct xpin_node *);
    const char         *(*qname)(struct xpin_node *);
    /* attributes: */
    int                 (*attr_count)(struct xpin_node *);
    /* namespace support: name is qname (ie, incl prefix): */
    const char         *(*attr_value)(struct xpin_node *, const char *name);
    struct xpin_attr  **(*attributes)(struct xpin_node *);
    /* node content: */
    struct xpin_text   *(*text)(struct xpin_node *);
    /* link to parent node: */
    struct xpin_node   *(*parent)(struct xpin_node *);
    /* content manipulation methods (convenience to underlying TEXT): */
    const char         *(*content)(struct xpin_node *);
    size_t              (*content_length)(struct xpin_node *);
    const char         *(*content_lower)(struct xpin_node *);
    const char         *(*content_upper)(struct xpin_node *);
    const char         *(*content_ltrim)(struct xpin_node *);
    const char         *(*content_rtrim)(struct xpin_node *);
    const char         *(*content_trim)(struct xpin_node *);
    const char         *(*content_chomp)(struct xpin_node *);
    const char         *(*content_normalize)(struct xpin_node *);
    const char         *(*content_append)(struct xpin_node *, const char *);
    const char         *(*content_prepend)(struct xpin_node *, const char *);
    /* append content to parent: */
    const char         *(*content_merge)(struct xpin_node *);
};
typedef struct xpin_node*  Xpin_Node;


/* ATTRIBUTES: */
struct xpin_attr {
    /* if namespace, name is qname */
    const char    *name;
    /* namespace support: */
    const char    *local_name;
    const char    *ns_uri;
    const char    *ns_prefix;
    const char    *value;    
};
/* typedef for an _array_ of node attributes: */
typedef struct xpin_attr**  Xpin_Attributes;


/* PIs (processing instructions): */
struct xpin_pi {
    /* opaque internal representation: */
    struct xpin_pi_  *pi;
    /* public methods: */
    const char      *(*target)(struct xpin_pi *);
    const char      *(*data)(struct xpin_pi *);
};
typedef struct xpin_pi*  Xpin_PI;


/* error object: */
struct xpin_error {
    /* opaque internal representation: */
    struct xpin_error_  *error;
    /* public methods: */
    int                 (*line)(struct xpin_error *);
    int                 (*column)(struct xpin_error *);
    int                 (*code)(struct xpin_error *);
    const char         *(*message)(struct xpin_error *);
};
typedef struct xpin_error*  Xpin_Error;

/* XpinDoc: */
struct xpindoc {
    /* opaque internal representation: */
    struct xpin    *xpin;
    /* public methods: */
    /* parse control: */
    void    (*set_mydata)(struct xpindoc *, void *);
    int     (*set_datamode)(struct xpindoc *, int datamode);
    int     (*parse_stream)(struct xpindoc *, FILE *);
    int     (*parse_buffer)(struct xpindoc *, char *);
    /* namespace processing on? */
    int     (*ns_parser)(struct xpindoc *);
    /* ignore chardata: */
    void    (*ignore_on)(struct xpindoc *);
    void    (*ignore_off)(struct xpindoc *);
    int     (*is_ignored)(struct xpindoc *);
    /* user halt support: */
    void    (*halt)(struct xpindoc *);
    int     (*is_halted)(struct xpindoc *);
    /* parse stopped by error: */
    int     (*is_croaked)(struct xpindoc *);
    /* handler installation: */
    int     (*set_handler)
                (
                    struct xpindoc *,
                    const char *,
                    void (*)(struct xpindoc *)
                );
    /* filter installation: */
    int     (*set_filter)
                (
                    struct xpindoc *,
                    void (*)(struct xpindoc *X, const char *data, int len)
                );
                        
    /* accessors: */
    void                 *(*mydata)(struct xpindoc *);
    int                   (*datamode)(struct xpindoc *);
    struct xpin_event    *(*event)(struct xpindoc *);
    struct xpin_node     *(*node)(struct xpindoc *);
    struct xpin_pi       *(*pi)(struct xpindoc *);
    struct xpin_error    *(*error)(struct xpindoc *);
    /* convenience methods accessing current event: */
    enum xpin_event_type  (*event_type)(struct xpindoc *);
    /* convenience methods accessing current node: */
    int                   (*depth)(struct xpindoc *);
    const char           *(*tagname)(struct xpindoc *);
    const char           *(*tagpath)(struct xpindoc *);
    const char           *(*content)(struct xpindoc *);
    struct xpin_text     *(*node_text)(struct xpindoc *);
    /* convenience methods accessing parser WRITER object: */
    int       (*put_char)(struct xpindoc *X, char c);
    int       (*put_string)(struct xpindoc *X, const char *string);
    int       (*put_chardata)
                  (
                      struct xpindoc *X,
                      const char     *data,
                      int             len
                   );
    int       (*printf)(struct xpindoc *X, const char *format, ...);
    /* convenience methods accessing WRITER output control: */
    int                   (*push_output)
                              (
                                  struct xpindoc *X,
                                  void           *dest,
                                  int             dest_type
                              );
    void                 *(*pop_output)(struct xpindoc *X);
    /* destructor: */
    void                  (*free)(struct xpindoc *);
};
typedef struct xpindoc*  XpinDoc;

/* XpinDoc object creator: */
extern XpinDoc  new_xpindoc(int flags);

/* typedef for signature of user-defined handler function: */
typedef void (*Xpin_Handler)(struct xpindoc *);

/* typedef for signature of user-defined filter function: */
typedef void (*Xpin_Filter)
    (
        struct xpindoc     *X,
        const char         *data,
        int                 len
    );

#endif  /* XPINDOC_H */
